"""
The module contains a set of auxiliary functions facilitating the tight-binding computations
"""
from __future__ import print_function
from __future__ import absolute_import
from itertools import product
import numpy as np
import yaml


def accum(accmap, input, func=None, size=None, fill_value=0, dtype=None):
    """An accumulation function similar to Matlab's `accumarray` function.

    Parameters
    ----------
    accmap : ndarray
        This is the "accumulation map".  It maps input (i.e. indices into
        `a`) to their destination in the output array.  The first `a.ndim`
        dimensions of `accmap` must be the same as `a.shape`.  That is,
        `accmap.shape[:a.ndim]` must equal `a.shape`.  For example, if `a`
        has shape (15,4), then `accmap.shape[:2]` must equal (15,4).  In this
        case `accmap[i,j]` gives the index into the output array where
        element (i,j) of `a` is to be accumulated.  If the output is, say,
        a 2D, then `accmap` must have shape (15,4,2).  The value in the
        last dimension give indices into the output array. If the output is
        1D, then the shape of `accmap` can be either (15,4) or (15,4,1)
    input : ndarray
        The input data to be accumulated.
    func : callable or None
        The accumulation function.  The function will be passed a list
        of values from `a` to be accumulated.
        If None, numpy.sum is assumed. (Default value = None)
    size : ndarray or None
        The size of the output array.  If None, the size will be determined
        from `accmap`. (Default value = None)
    fill_value : scalar
        The default value for elements of the output array.
    dtype : numpy data type, or None
        The data type of the output array.  If None, the data type of
        `a` is used. (Default value = None)

    Returns
    -------
    out : ndarray
        The accumulated results.
        The shape of `out` is `size` if `size` is given.  Otherwise the
        shape is determined by the (lexicographically) largest indices of
        the output found in `accmap`.

    Examples
    --------
    >>> from numpy import array, prod
    >>> a = array([[1, 2, 3], [4, -1, 6], [-1, 8, 9]])
    >>> a
    array([[ 1,  2,  3],
           [ 4, -1,  6],
           [-1,  8,  9]])
    >>> # Sum the diagonals.
    >>> accmap = array([[0, 1, 2], [2, 0, 1], [1, 2, 0]])
    >>> s = accum(accmap,a)
    >>> s
    array([ 9,  7, 15])
    >>> # A 2D output, from sub-arrays with shapes and positions like this:
    >>> # [ (2,2) (2,1)]
    >>> # [ (1,2) (1,1)]
    >>> accmap = array([[[0,0],[0,0],[0,1]],[[0,0],[0,0],[0,1]],[[1,0],[1,0],[1,1]],])
    >>> # Accumulate using a product.
    >>> accum(accmap,a,func=prod,dtype=float)
    array([[-8., 18.],
           [-8.,  9.]])
    >>> # Same accmap, but create an array of lists of values.
    >>> accum(accmap,a,func=lambda x: x,dtype='O')
    array([[list([1, 2, 4, -1]), list([3, 6])],
           [list([-1, 8]), list([9])]], dtype=object)
    """

    # Check for bad arguments and handle the defaults.
    if accmap.shape[:input.ndim] != input.shape:
        raise ValueError("The initial dimensions of accmap must be the same as a.shape")
    if func is None:
        func = np.sum
    if dtype is None:
        dtype = input.dtype
    if accmap.shape == input.shape:
        accmap = np.expand_dims(accmap, -1)
    adims = tuple(range(input.ndim))
    if size is None:
        size = 1 + np.squeeze(np.apply_over_axes(np.max, accmap, axes=adims))
    size = np.atleast_1d(size)

    # Create an array of python lists of values.
    vals = np.empty(size, dtype='O')
    for s in product(*[range(k) for k in size]):
        vals[s] = []
    for s in product(*[range(k) for k in input.shape]):
        indx = tuple(accmap[s])
        val = input[s]
        vals[indx].append(val)

    # Create the output array.
    out = np.empty(size, dtype=dtype)
    for s in product(*[range(k) for k in size]):
        if vals[s] == []:
            out[s] = fill_value
        else:
            out[s] = func(vals[s])

    return out


def xyz2np(xyz):
    """Transforms xyz-file formatted string to lists of atomic labels and coordinates

    Parameters
    ----------
    xyz :
        xyz-formatted string

    Returns
    -------
    list, list
        list of labels and list of coordinates

    """

    xyz = xyz.splitlines()
    num_of_atoms = int(xyz[0])
    ans = np.zeros((num_of_atoms, 3))
    j = 0
    atoms = []
    unique_labels = dict()

    for line in xyz[2:]:
        if len(line.strip()) > 0:
            temp = line.split()
            label = ''.join([i for i in temp[0] if not i.isdigit()])

            try:
                unique_labels[label] += 1
                temp[0] = label + str(unique_labels[label])
            except KeyError:
                temp[0] = label + '1'
                unique_labels[label] = 1

            atoms.append(temp[0])
            ans[j, 0] = float(temp[1])
            ans[j, 1] = float(temp[2])
            ans[j, 2] = float(temp[3])
            j += 1

    return atoms, ans


def count_species(list_of_labels):
    """From the list of labels creates a dictionary where the keys represents labels and
    values are numbers of their repetitions in the list

    Parameters
    ----------
    list_of_labels :
        return:

    Returns
    -------

    """
    counter = {}

    for item in list_of_labels:

        key = ''.join([i for i in item if not i.isdigit()])

        try:
            counter[key] += 1
        except KeyError:
            counter[key] = 1

    return counter


def get_k_coords(special_points, num_of_points, label):
    """Generates a array of the coordinates in the k-space from the set of
    high-symmetry points and number of nodes between them

    Parameters
    ----------
    special_points :
        list of labels for high-symmetry points
    num_of_points :
        list of node numbers in each section of the path in the k-space
    label :
        chemical element

    Returns
    -------
    numpy.ndarray
        array of coordinates in k-space

    """

    from nanonet.tb.special_points import SPECIAL_K_POINTS_BI, SPECIAL_K_POINTS_SI

    if isinstance(label, str):
        if label == 'Bi':
            SPECIAL_K_POINTS = SPECIAL_K_POINTS_BI
        if label == 'Si':
            SPECIAL_K_POINTS = SPECIAL_K_POINTS_SI
    else:
        SPECIAL_K_POINTS = label

    k_vectors = np.zeros((sum(num_of_points), 3))
    offset = 0

    for j in range(len(num_of_points)):
        sequence1 = np.linspace(SPECIAL_K_POINTS[special_points[j]][0],
                                SPECIAL_K_POINTS[special_points[j + 1]][0], num_of_points[j])
        sequence2 = np.linspace(SPECIAL_K_POINTS[special_points[j]][1],
                                SPECIAL_K_POINTS[special_points[j + 1]][1], num_of_points[j])
        sequence3 = np.linspace(SPECIAL_K_POINTS[special_points[j]][2],
                                SPECIAL_K_POINTS[special_points[j + 1]][2], num_of_points[j])

        k_vectors[offset:offset + num_of_points[j], :] = \
            np.vstack((sequence1, sequence2, sequence3)).T

        offset += num_of_points[j]

    return k_vectors


def dict2xyz(input_data):
    """

    Parameters
    ----------
    input_data :
        return:

    Returns
    -------

    """

    if not isinstance(input_data, dict):
        return input_data

    output = str(input_data['num_atoms']) + '\n'
    output += str(input_data['title']) + '\n'

    for j in range(input_data['num_atoms']):
        output += list(input_data['atoms'][j].keys())[0] + \
                  "    " + str(list(input_data['atoms'][j].values())[0][0]) + \
                  "    " + str(list(input_data['atoms'][j].values())[0][1]) + \
                  "    " + str(list(input_data['atoms'][j].values())[0][2]) + "\n"

    return output


def yaml_parser(input_data):
    """

    Parameters
    ----------
    input_data :
        return:

    Returns
    -------

    """

    output = None

    if input_data.lower().endswith(('.yml', '.yaml')):
        with open(input_data, 'r') as stream:
            try:
                output = yaml.safe_load(stream)
            except yaml.YAMLError as exc:
                print(exc)
    else:
        try:
            output = yaml.safe_load(input_data)
        except yaml.YAMLError as exc:
            print(exc)

    output['primitive_cell'] = np.array(output['primitive_cell']) * output['lattice_constant']

    return output


def print_table(myDict, colList=None, sep='\uFFFA'):
    """Pretty print a list of dictionaries (myDict) as a dynamically sized table.
    If column names (colList) aren't specified, they will show in random order.
    sep: row separator. Ex: sep='\n' on Linux. Default: dummy to not split line.
    Author: Thierry Husson - Use it as you want but don't blame me.

    Parameters
    ----------
    myDict :
        
    colList :
         (Default value = None)
    sep :
         (Default value = '\uFFFA')

    Returns
    -------

    """

    if not colList:
        colList = list(myDict[0].keys() if myDict else [])

    myList = [colList]  # 1st row = header

    for item in myDict:
        myList.append([str(item[col]) for col in colList])

    colSize = [max(map(len, (sep.join(col)).split(sep))) for col in zip(*myList)]

    formatStr = ' | '.join(["{{:<{}}}".format(i) for i in colSize])
    line = formatStr.replace(' | ', '-+-').format(*['-' * i for i in colSize])
    item = myList.pop(0)
    lineDone = False

    out = "\n"

    while myList:
        if all(not i for i in item):
            item = myList.pop(0)
            if line and (sep != '\uFFFA' or not lineDone):
                out += line
                out += "\n"
                lineDone = True

        row = [i.split(sep, 1) for i in item]
        out += formatStr.format(*[i[0] for i in row])
        out += "\n"
        item = [i[1] if len(i) > 1 else '' for i in row]

    out += line
    out += "\n"

    return out


def print_dict(dictionary):
    """

    Parameters
    ----------
    dictionary :
        

    Returns
    -------

    """
    out = "{:<18} {:<15} \n".format('Label', 'Coordinates')
    for key, value in dictionary.items():
        out += "{:<18} {:<15} \n".format(key, str(value))

    return out


# def split_into_subblocks(h_0, h_l, h_r):
#     """
#     Split Hamiltonian matrix and coupling matrices into subblocks
#
#     :param h_0:                     Hamiltonian matrix
#     :param h_l:                     left inter-cell coupling matrices
#     :param h_r:                     right inter-cell coupling matrices
#     :return h_0_s, h_l_s, h_r_s:    lists of subblocks
#     """
#
#     def find_nonzero_lines(mat, order):
#
#         if order == 'top':
#             line = mat.shape[0]
#             while line > 0:
#                 if np.count_nonzero(mat[line - 1, :]) == 0:
#                     line -= 1
#                 else:
#                     break
#         elif order == 'bottom':
#             line = -1
#             while line < mat.shape[0] - 1:
#                 if np.count_nonzero(mat[line + 1, :]) == 0:
#                     line += 1
#                 else:
#                     line = mat.shape[0] - (line + 1)
#                     break
#         elif order == 'left':
#             line = mat.shape[1]
#             while line > 0:
#                 if np.count_nonzero(mat[:, line - 1]) == 0:
#                     line -= 1
#                 else:
#                     break
#         elif order == 'right':
#             line = -1
#             while line < mat.shape[1] - 1:
#                 if np.count_nonzero(mat[:, line + 1]) == 0:
#                     line += 1
#                 else:
#                     line = mat.shape[1] - (line + 1)
#                     break
#         else:
#             raise ValueError('Wrong value of the parameter order')
#
#         return line
#
#     h_0_s = []
#     h_l_s = []
#     h_r_s = []
#
#     if isinstance(h_l, np.ndarray) and isinstance(h_r, np.ndarray):
#         h_r_h = find_nonzero_lines(h_r, 'bottom')
#         h_r_v = find_nonzero_lines(h_r[-h_r_h:, :], 'left')
#         h_l_h = find_nonzero_lines(h_l, 'top')
#         h_l_v = find_nonzero_lines(h_l[:h_l_h, :], 'right')
#
#     if isinstance(h_l, int) and isinstance(h_r, int):
#         h_l_h = h_l
#         h_r_v = h_l
#         h_r_h = h_r
#         h_l_v = h_r
#
#     edge, edge1 = compute_edge(h_0)
#
#     left_block = max(h_l_h, h_r_v)
#     right_block = max(h_r_h, h_l_v)
#
#     from nanonet.tb.sorting_algorithms import split_matrix_0, cut_in_blocks, split_matrix
#
#     blocks = split_matrix_0(h_0, left=left_block, right=right_block)
#     h_0_s, h_l_s, h_r_s = cut_in_blocks(h_0, blocks)
#
#     # blocks = compute_blocks(left_block, right_block, edge, edge1)
#     # j1 = 0
#     #
#     # for j, block in enumerate(blocks):
#     #     h_0_s.append(h_0[j1:block + j1, j1:block + j1])
#     #     if j < len(blocks) - 1:
#     #         h_l_s.append(h_0[block + j1:block + j1 + blocks[j + 1], j1:block + j1])
#     #         h_r_s.append(h_0[j1:block + j1, j1 + block:j1 + block + blocks[j + 1]])
#     #     j1 += block
#
#     return h_0_s, h_l_s, h_r_s, blocks


def compute_edge(mat):
    """

    Parameters
    ----------
    mat :
        

    Returns
    -------

    """
    # First get some statistics
    row, col = np.where(mat != 0.0)  # Output rows and columns of all non-zero elements.

    # Clever use of accumarray:
    outeredge = accum(row, col, np.max) + 1
    outeredge[0] = max(0, outeredge[0])
    outeredge = np.maximum.accumulate(outeredge)

    outeredge1 = accum(np.max(row) - row[::-1], np.max(row) - col[::-1], np.max) + 1
    outeredge1[0] = max(0, outeredge1[0])
    outeredge1 = np.maximum.accumulate(outeredge1)

    return outeredge, outeredge1


def blocksandborders_constrained(left_block, right_block, edge, edge1):
    """A version of blocksandborders with constraints - periodic boundary conditions.

    Parameters
    ----------
    mat :
        input matrix
    left_block :
        left block constrained minimal size
    right_block :
        right block constrained minimal size
    edge :
        edge of sparsity pattern
    edge1 :
        conjugate edge of sparsity pattern

    Returns
    -------
    type
        array of diagonal block sizes

    """

    size = len(edge)
    left_block = max(1, left_block)
    right_block = max(1, right_block)

    if left_block + right_block < size:                               # if blocks do not overlap

        new_left_block = edge[left_block - 1] - left_block
        new_right_block = edge1[right_block - 1] - right_block
        #
        # new_right_block = np.max(np.argwhere(np.abs(edge - (size - right_block)) -
        #                                      np.min(np.abs(edge - (size - right_block))) == 0)) + 1
        # new_right_block = size - new_right_block - right_block

        if left_block + new_left_block <= size - right_block and\
                size - right_block - new_right_block >= left_block:    # spacing between blocks is sufficient

            blocks = blocksandborders_constrained(new_left_block,
                                                  new_right_block,
                                                  edge[left_block:-right_block] - left_block,
                                                  edge1[right_block:-left_block] - right_block)

            return [left_block] + blocks + [right_block]
        else:
            if new_left_block > new_right_block:
                return [left_block] + [size - left_block]
            else:
                return [size - right_block] + [right_block]

    elif left_block + right_block == size:                            # sum of blocks equal to the matrix size
        return [left_block] + [right_block]
    else:                                                             # blocks overlap
        return [size]


def argsort(seq):
    """

    Parameters
    ----------
    seq :
        

    Returns
    -------

    """
    # http://stackoverflow.com/questions/3071415/efficient-method-to-calculate-the-rank-vector-of-a-list-in-python
    return sorted(range(len(seq)), key=seq.__getitem__)


def shift(mat):
    """

    Parameters
    ----------
    mat :
        

    Returns
    -------

    """

    ans = np.zeros(mat.shape, dtype=np.int)

    cut = mat.shape[0] // 2

    ans[:cut] = mat[cut:]
    ans[cut:] = mat[:cut]

    return ans


def bandwidth1(mat):
    """

    Parameters
    ----------
    mat :
        

    Returns
    -------

    """

    j = 0

    while np.count_nonzero((np.diag(mat, mat.shape[0] - j - 1))) == 0 and j < mat.shape[0]:
        j += 1

    return mat.shape[0] - j - 1


def bandwidth(mat):
    """

    Parameters
    ----------
    mat :
        

    Returns
    -------

    """

    ans = 0

    for j in range(1, mat.shape[0]):
        if np.count_nonzero((np.diag(mat, j))) > 0:
            ans = j

    return ans


# Helper function to store the inroder traversal of a tree
def storeInorder(root, inorder):
    """

    Parameters
    ----------
    root :
        
    inorder :
        

    Returns
    -------

    """
    # Base Case
    if root is None:
        return

        # First store the left subtree
    storeInorder(root.left, inorder)

    # Copy the root's data
    inorder.append(root.data)

    # Finally store the right subtree
    storeInorder(root.right, inorder)


# A helper funtion to count nodes in a binary tree
def countNodes(root):
    """

    Parameters
    ----------
    root :
        

    Returns
    -------

    """
    if root is None:
        return 0

    return countNodes(root.lesser) + countNodes(root.greater) + 1


def is_in_coords(coord, coords):
    """

    Parameters
    ----------
    coord :
        
    coords :
        

    Returns
    -------

    """

    ans = False

    for xyz in list(coords):
        ans += (np.linalg.norm(coord - xyz) < 0.01)

    return ans


def fd(energy, ef, temp):
    """
    Fermi-Dirac function

    Parameters
    ----------
    energy :

    ef :

    temp :


    Returns
    -------

    """
    kb = 8.61733e-5  # Boltzmann constant in eV
    return 1.0 / (1.0 + np.exp((energy - ef) / (kb * temp)))



from __future__ import absolute_import
from pkg_resources import get_distribution

__version__ = get_distribution('nano-net').version
__author__ = "Mike Klymenko"
__email__ = "mike.klymenko@rmit.edu.au"

from .greens_functions import surface_greens_function
from .recursive_greens_functions import recursive_gf


"""This module contains a set of functions facilitating computations of
the block-tridiagonal structure of a band matrix.
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from itertools import product
import math
import scipy


def accum(accmap, input, func=None, size=None, fill_value=0, dtype=None):
    """An accumulation function similar to Matlab's `accumarray` function.

    Parameters
    ----------
    accmap : ndarray
        This is the "accumulation map".  It maps input (i.e. indices into
        `a`) to their destination in the output array.  The first `a.ndim`
        dimensions of `accmap` must be the same as `a.shape`.  That is,
        `accmap.shape[:a.ndim]` must equal `a.shape`.  For example, if `a`
        has shape (15,4), then `accmap.shape[:2]` must equal (15,4).  In this
        case `accmap[i,j]` gives the index into the output array where
        element (i,j) of `a` is to be accumulated.  If the output is, say,
        a 2D, then `accmap` must have shape (15,4,2).  The value in the
        last dimension give indices into the output array. If the output is
        1D, then the shape of `accmap` can be either (15,4) or (15,4,1)
    input : ndarray
        The input data to be accumulated.
    func : callable or None
        The accumulation function.  The function will be passed a list
        of values from `a` to be accumulated.
        If None, numpy.sum is assumed. (Default value = None)
    size : ndarray or None
        The size of the output array.  If None, the size will be determined
        from `accmap`. (Default value = None)
    fill_value : scalar
        The default value for elements of the output array.
    dtype : numpy data type, or None
        The data type of the output array.  If None, the data type of
        `a` is used. (Default value = None)

    Returns
    -------

    
    """

    # Check for bad arguments and handle the defaults.
    if accmap.shape[:input.ndim] != input.shape:
        raise ValueError("The initial dimensions of accmap must be the same as a.shape")
    if func is None:
        func = np.sum
    if dtype is None:
        dtype = input.dtype
    if accmap.shape == input.shape:
        accmap = np.expand_dims(accmap, -1)
    adims = tuple(range(input.ndim))
    if size is None:
        size = 1 + np.squeeze(np.apply_over_axes(np.max, accmap, axes=adims))
    size = np.atleast_1d(size)

    # Create an array of python lists of values.
    vals = np.empty(size, dtype='O')
    for s in product(*[range(k) for k in size]):
        vals[s] = []
    for s in product(*[range(k) for k in input.shape]):
        indx = tuple(accmap[s])
        val = input[s]
        vals[indx].append(val)

    # Create the output array.
    out = np.empty(size, dtype=dtype)
    for s in product(*[range(k) for k in size]):
        if vals[s] == []:
            out[s] = fill_value
        else:
            out[s] = func(vals[s])

    return out


def cut_in_blocks(h_0, blocks):
    """Cut a matrix into diagonal, upper-diagonal and lower-diagonal blocks
    if sizes of the diagonal blocks are specified.

    Parameters
    ----------
    h_0 : ndarray
        Input matrix
    blocks : ndarray(dtype=int)
        Sizes of diagonal blocks

    Returns
    -------
    h_0_s, h_l_s, h_r_s : ndarray
        List of diagonal matrices,
        list of lower-diagonal matrices and
        list of upper-diagonal matrices.
        Note that if the size of the list h_0_s is N,
        the sizes of h_l_s, h_r_s are N-1.

    Examples
    --------
    >>> import numpy as np
    >>> from nanonet.tb.block_tridiagonalization import cut_in_blocks
    >>> a = np.array([[1, 1, 0, 0], [1, 1, 1, 0], [0, 1, 1, 1], [0, 0, 1, 1]])
    >>> a
    array([[1, 1, 0, 0],
           [1, 1, 1, 0],
           [0, 1, 1, 1],
           [0, 0, 1, 1]])
    >>> # Sum the diagonals.
    >>> blocks = [2, 2]
    >>> blocks
    [2, 2]
    >>> h0, h1, h2 = cut_in_blocks(a, blocks)
    >>> h0
    [array([[1, 1],
           [1, 1]]), array([[1, 1],
           [1, 1]])]
    >>> h1
    [array([[0, 1],
           [0, 0]])]
    >>> h2
    [array([[0, 0],
           [1, 0]])]
    """

    j1 = 0

    h_0_s = []
    h_l_s = []
    h_r_s = []

    for j, block in enumerate(blocks):
        h_0_s.append(h_0[j1:block + j1, j1:block + j1])
        if j < len(blocks) - 1:
            h_l_s.append(h_0[block + j1:block + j1 + blocks[j + 1], j1:block + j1])
            h_r_s.append(h_0[j1:block + j1, j1 + block:j1 + block + blocks[j + 1]])
        j1 += block

    return h_0_s, h_l_s, h_r_s


def find_optimal_cut(edge, edge1, left, right):
    """Computes the index corresponding to the optimal cut such that applying
    the function compute_blocks() to the sub-blocks defined by the cut reduces
    the cost function comparing to the case when the function compute_blocks() is
    applied to the whole matrix. If cutting point can not be find, the algorithm returns
    the result from the function compute_blocks().

    Parameters
    ----------
    edge : ndarray
        sparsity pattern profile of the matrix
    edge1 : ndarray
        conjugated sparsity pattern profile of the matrix
    left : int
        size of the leftmost diagonal block
    right : int
        size of the rightmost diagonal block

    Returns
    -------

    
    """

    unique_indices = np.arange(left, len(edge) - right + 1)
    blocks = []
    seps = []
    sizes = []
    metric = []
    size = len(edge)

    for j1, item1 in enumerate(unique_indices):
        seps.append(item1)
        item2 = size - item1

        # print(item1, item2)
        # print(item1)

        edge_1 = edge[:item1]
        edge_2 = (edge1 - np.arange(len(edge1)))[item2:] + np.arange(item1)

        edge_3 = edge1[:item2]
        edge_4 = (edge - np.arange(len(edge)))[item1:] + np.arange(item2)

        block1 = compute_blocks(left, (edge1 - np.arange(len(edge)))[item2],
                                edge_1, edge_2)

        block2 = compute_blocks(right, (edge - np.arange(len(edge1)))[item1],
                                edge_3, edge_4)

        block = block1 + block2[::-1]
        blocks.append(block)
        metric.append(np.sum(np.array(block) ** 3))
        sizes.append((block1[-1], block2[-1]))

    if len(metric) == 0:
        return [left, right], np.nan, 0, 0
    else:

        best = np.argmin(np.array(metric))

        blocks = blocks[best]
        blocks = [item for item in blocks if item != 0]

        sep = seps[best]

        right_block, left_block = sizes[best]

        return blocks, sep, right_block, left_block


def compute_blocks_optimized(edge, edge1, left=1, right=1):
    """Computes optimal sizes of diagonal blocks of a matrix whose
    sparsity pattern is defined by the sparsity pattern profiles edge and edge1.
    This function is based on the algorithm which uses defined above function
    find_optimal_cut() to subdivide the problem into sub-problems in a optimal way
    according to some cost function.

    Parameters
    ----------
    edge : ndarray
        sparsity pattern profile of the matrix
    edge1 : ndarray
        conjugated sparsity pattern profile of the matrix
    left : int
        size of the leftmost diagonal block (constrained) (Default value = 1)
    right : int
        size of the rightmost diagonal block (constrained) (Default value = 1)

    Returns
    -------

    
    """

    blocks, sep, right_block, left_block = find_optimal_cut(edge, edge1, left=left, right=right)
    flag = False

    if not math.isnan(sep):

        # print(left, right_block, sep)

        if left + right_block < sep:

            edge_1 = edge[:sep]
            # edge_1[edge_1 > sep] = sep
            edge_2 = (edge1 - np.arange(len(edge1)))[-sep:] + np.arange(sep)

            blocks1 = compute_blocks_optimized(edge_1, edge_2, left=left, right=right_block)

        elif left + right_block == sep:

            blocks1 = [left, right_block]
        else:

            flag = True

        # print(left_block, right, len(edge) - sep)

        if right + left_block < len(edge) - sep:

            edge_3 = (edge - np.arange(len(edge)))[sep:] + np.arange(len(edge) - sep)
            edge_4 = edge1[:-sep]
            # edge_4[edge_4 > len(edge) - sep] = len(edge) - sep

            blocks2 = compute_blocks_optimized(edge_3, edge_4, left=left_block, right=right)

        elif right + left_block == len(edge) - sep:
            blocks2 = [left_block, right]
        else:
            flag = True

        if flag:
            return blocks
        else:
            blocks = blocks1 + blocks2

            return blocks


def find_nonzero_lines(mat, order):
    """

    Parameters
    ----------
    mat :
        
    order :
        

    Returns
    -------

    """

    if scipy.sparse.issparse(mat):
        lines = _find_nonzero_lines_sparse(mat, order)
    else:
        lines = _find_nonzero_lines(mat, order)

    if lines == max(mat.shape[0], mat.shape[1]) - 1:
        lines = 1
    if lines == 0:
        lines = 1

    return lines


def _find_nonzero_lines(mat, order):
    """

    Parameters
    ----------
    mat :
        
    order :
        

    Returns
    -------

    """
    if order == 'top':
        line = mat.shape[0]
        while line > 0:
            if np.count_nonzero(mat[line - 1, :]) == 0:
                line -= 1
            else:
                break
    elif order == 'bottom':
        line = -1
        while line < mat.shape[0] - 1:
            if np.count_nonzero(mat[line + 1, :]) == 0:
                line += 1
            else:
                line = mat.shape[0] - (line + 1)
                break
    elif order == 'left':
        line = mat.shape[1]
        while line > 0:
            if np.count_nonzero(mat[:, line - 1]) == 0:
                line -= 1
            else:
                break
    elif order == 'right':
        line = -1
        while line < mat.shape[1] - 1:
            if np.count_nonzero(mat[:, line + 1]) == 0:
                line += 1
            else:
                line = mat.shape[1] - (line + 1)
                break
    else:
        raise ValueError('Wrong value of the parameter order')

    return line


def _find_nonzero_lines_sparse(mat, order):
    """

    Parameters
    ----------
    mat :
        
    order :
        

    Returns
    -------

    """
    if order == 'top':
        line = mat.shape[0]
        while line > 0:
            if np.count_nonzero(mat[line - 1, :].todense()) == 0:
                line -= 1
            else:
                break
    elif order == 'bottom':
        line = -1
        while line < mat.shape[0] - 1:
            if np.count_nonzero(mat[line + 1, :].todense()) == 0:
                line += 1
            else:
                line = mat.shape[0] - (line + 1)
                break
    elif order == 'left':
        line = mat.shape[1]
        while line > 0:
            if np.count_nonzero(mat[:, line - 1].todense()) == 0:
                line -= 1
            else:
                break
    elif order == 'right':
        line = -1
        while line < mat.shape[1] - 1:
            if np.count_nonzero(mat[:, line + 1].todense()) == 0:
                line += 1
            else:
                line = mat.shape[1] - (line + 1)
                break
    else:
        raise ValueError('Wrong value of the parameter order')

    return line


def split_into_subblocks_optimized(h_0, left=1, right=1):
    """

    Parameters
    ----------
    h_0 :
        param left:
    right :
        return: (Default value = 1)
    left :
         (Default value = 1)

    Returns
    -------

    """

    if not (isinstance(left, int) and isinstance(right, int)):
        h_r_h = find_nonzero_lines(right, 'bottom')
        h_r_v = find_nonzero_lines(right[-h_r_h:, :], 'left')
        h_l_h = find_nonzero_lines(left, 'top')
        h_l_v = find_nonzero_lines(left[:h_l_h, :], 'right')
        left = max(h_l_h, h_r_v)
        right = max(h_r_h, h_l_v)

    if left + right > h_0.shape[0]:
        return [h_0.shape[0]]
    else:
        edge, edge1 = compute_edge(h_0)
        return compute_blocks_optimized(edge, edge1, left=left, right=right)


def split_into_subblocks(h_0, h_l, h_r):
    """Split Hamiltonian matrix and coupling matrices into subblocks

    Parameters
    ----------
    h_0 :
        Hamiltonian matrix
    h_l :
        left inter-cell coupling matrices
    h_r :
        right inter-cell coupling matrices
        :return h_0_s, h_l_s, h_r_s:    lists of subblocks

    Returns
    -------

    """

    if isinstance(h_l, np.ndarray) and isinstance(h_r, np.ndarray):
        h_r_h = find_nonzero_lines(h_r, 'bottom')
        h_r_v = find_nonzero_lines(h_r[-h_r_h:, :], 'left')
        h_l_h = find_nonzero_lines(h_l, 'top')
        h_l_v = find_nonzero_lines(h_l[:h_l_h, :], 'right')
        left_block = max(h_l_h, h_r_v)
        right_block = max(h_r_h, h_l_v)
    elif isinstance(h_l, int) and isinstance(h_r, int):
        left_block = h_l
        right_block = h_r
    else:
        raise TypeError

    edge, edge1 = compute_edge(h_0)

    blocks = compute_blocks(left_block, right_block, edge, edge1)

    return blocks


def compute_edge(mat):
    """Computes edges of the sparsity pattern of a matrix.

    Parameters
    ----------
    mat : ndarray
        Input matrix

    Returns
    -------
    edge : ndarray
        edge of the sparsity pattern
    edge1 : ndarray
        conjugate edge of the sparsity pattern

    Examples
    --------
    >>> import numpy as np
    >>> from nanonet.tb.block_tridiagonalization import compute_edge
    >>> input_matrix = np.array([[1, 1, 0, 0], [1, 1, 1, 0], [0, 1, 1, 1], [0, 0, 1, 1]])
    >>> input_matrix
    array([[1, 1, 0, 0],
           [1, 1, 1, 0],
           [0, 1, 1, 1],
           [0, 0, 1, 1]])
    >>> e1, e2 = compute_edge(input_matrix)
    >>> e1
    array([2, 3, 4, 4])
    >>> e2
    array([2, 3, 4, 4])
    >>> input_matrix = np.array([[1, 0, 0, 0], [0, 1, 1, 0], [0, 1, 1, 1], [0, 0, 1, 1]])
    >>> input_matrix
    array([[1, 0, 0, 0],
           [0, 1, 1, 0],
           [0, 1, 1, 1],
           [0, 0, 1, 1]])
    >>> e1, e2 = compute_edge(input_matrix)
    >>> e1
    array([1, 3, 4, 4])
    >>> e2
    array([2, 3, 3, 4])
    """

    # First get some statistics
    if isinstance(mat, scipy.sparse.lil.lil_matrix):
        row, col = mat.nonzero()
    else:
        row, col = np.where(mat != 0.0)  # Output rows and columns of all non-zero elements.

    # Clever use of accumarray:
    edge = accum(row, col, np.max) + 1
    edge[0] = max(0, edge[0])
    edge = np.maximum.accumulate(edge)

    edge1 = accum(np.max(row) - row[::-1], np.max(row) - col[::-1], np.max) + 1
    edge1[0] = max(0, edge1[0])
    edge1 = np.maximum.accumulate(edge1)

    return edge, edge1


def compute_blocks(left_block, right_block, edge, edge1):
    """This is an implementation of the greedy algorithm for
     computing block-tridiagonal representation of a matrix.
     The information regarding the input matrix is represented
     by the sparsity patters edges, `edge` and `edge1`.

    Parameters
    ----------
    left_block : int
        a predefined size of the leftmost block
    right_block : int
        a predefined size of the rightmost block
    edge : ndarray
        edge of sparsity pattern
    edge1 : ndarray
        conjugate edge of sparsity pattern

    Returns
    -------
    ans : list
        

    Examples
    --------
    >>> import numpy as np
    >>> from nanonet.tb.block_tridiagonalization import compute_edge
    >>> input_matrix = np.array([[1, 1, 0, 0], [1, 1, 1, 0], [0, 1, 1, 1], [0, 0, 1, 1]])
    >>> input_matrix
    array([[1, 1, 0, 0],
           [1, 1, 1, 0],
           [0, 1, 1, 1],
           [0, 0, 1, 1]])
    >>> e1, e2 = compute_edge(input_matrix)
    >>> compute_blocks(1, 1, e1, e2)
    [1, 1, 1, 1]
    >>> input_matrix = np.array([[1, 1, 1, 0], [1, 1, 1, 0], [1, 1, 1, 1], [0, 0, 1, 1]])
    >>> input_matrix
    array([[1, 1, 1, 0],
           [1, 1, 1, 0],
           [1, 1, 1, 1],
           [0, 0, 1, 1]])
    >>> e1, e2 = compute_edge(input_matrix)
    >>> compute_blocks(1, 1, e1, e2)
    [1, 2, 1]
    >>> e1, e2 = compute_edge(input_matrix)
    >>> compute_blocks(2, 2, e1, e2)
    [2, 2]
    """

    size = len(edge)
    left_block = max(1, left_block)
    right_block = max(1, right_block)

    if left_block + right_block < size:  # if blocks do not overlap

        new_left_block = edge[left_block - 1] - left_block
        new_right_block = edge1[right_block - 1] - right_block
        #
        # new_right_block = np.max(np.argwhere(np.abs(edge - (size - right_block)) -
        #                                      np.min(np.abs(edge - (size - right_block))) == 0)) + 1
        # new_right_block = size - new_right_block - right_block

        if left_block + new_left_block <= size - right_block and \
                size - right_block - new_right_block >= left_block:  # spacing between blocks is sufficient

            blocks = compute_blocks(new_left_block,
                                    new_right_block,
                                    edge[left_block:-right_block] - left_block,
                                    edge1[right_block:-left_block] - right_block)

            return [left_block] + blocks + [right_block]
        else:
            if new_left_block > new_right_block:
                return [left_block] + [size - left_block]
            else:
                return [size - right_block] + [right_block]

    elif left_block + right_block == size:  # sum of blocks equal to the matrix size
        return [left_block] + [right_block]
    else:  # blocks overlap
        return [size]


def show_blocks(subblocks, input_mat):
    """This is a script for visualizing the sparsity pattern and
     a block-tridiagonal structure of a matrix.

    Parameters
    ----------
    subblocks :
        
    input_mat :
        

    Returns
    -------

    
    """

    cumsum = np.cumsum(np.array(subblocks))[:-1]
    cumsum = np.insert(cumsum, 0, 0)

    fig, ax = plt.subplots(1)
    plt.spy(input_mat, markersize=0.9, c='k')
    # plt.plot(edge)

    for jj in range(2):
        cumsum = cumsum + jj * input_mat.shape[0]

        if jj == 1:
            rect = Rectangle((input_mat.shape[0] - subblocks[-1] - 0.5, input_mat.shape[1] - 0.5),
                             subblocks[-1], subblocks[0],
                             linestyle='--',
                             linewidth=1.3,
                             edgecolor='b',
                             facecolor='none', zorder=200)
            ax.add_patch(rect)
            rect = Rectangle((input_mat.shape[0] - 0.5, input_mat.shape[1] - subblocks[-1] - 0.5),
                             subblocks[0], subblocks[-1],
                             linestyle='--',
                             linewidth=1.3,
                             edgecolor='g',
                             facecolor='none', zorder=200)
            ax.add_patch(rect)

        for j, item in enumerate(cumsum):
            if j < len(cumsum) - 1:
                rect = Rectangle((item - 0.5, cumsum[j + 1] - 0.5), subblocks[j], subblocks[j + 1],
                                 linewidth=1.3,
                                 edgecolor='b',
                                 facecolor='none', zorder=200)
                ax.add_patch(rect)
                rect = Rectangle((cumsum[j + 1] - 0.5, item - 0.5), subblocks[j + 1], subblocks[j],
                                 linewidth=1.3,
                                 edgecolor='g',
                                 facecolor='none', zorder=200)
                ax.add_patch(rect)
            rect = Rectangle((item - 0.5, item - 0.5), subblocks[j], subblocks[j],
                             linewidth=1.3,
                             edgecolor='r',
                             facecolor='none', zorder=200)
            ax.add_patch(rect)

    plt.xlim(input_mat.shape[0] - 0.5, -1.0)
    plt.ylim(-1.0, input_mat.shape[0] - 0.5)
    plt.axis('off')
    plt.show()


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
The module contains functions computing hopping parameters
with arbitrary rotations of atomic orbitals based on the table of
empirical diatomic couplings defined in the module params.
Computations are based mostly on analytical equations derived in
[A.V. Podolskiy and P. Vogl, Phys. Rev. B. 69, 233101 (2004)].
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division
import sys
import math
from nanonet.tb.constants import *
import warnings
from nanonet.tb import tb_params


def me_diatomic(bond, n, l_min, l_max, m, which_neighbour, overlap=False):
    """The function looks up into the table of parameters making a query parametrized by:

    Parameters
    ----------
    bond :
        a bond type represented by a list of atom labels
    n :
        combination of the principal quantum numbers of atoms
    l_min :
        min(l1, l2), where l1 and l2 are orbital quantum numbers of atoms
    l_max :
        max(l1, l2), where l1 and l2 are orbital quantum numbers of atoms
    m :
        symmetry of the electron wave function in the diatomic molecule
        takes values "sigma", "pi" and "delta"
    which_neighbour :

    overlap : bool
        A flag indicating that the overlap matrix element has to be computed
        

    Returns
    -------
    float
        numerical value of the corresponding tabular parameter

    """

    label = n[0] + ORBITAL_QN[l_min] + n[1] + ORBITAL_QN[l_max] + '_' + M_QN[m]


    if overlap:
        flag = 'OV_'
    else:
        flag = 'PARAMS_'

    try:
        if which_neighbour == 0:
            return getattr(sys.modules[tb_params.__name__], flag + bond)[label]
        elif which_neighbour == 100:
            return 0
        else:
            return getattr(sys.modules[tb_params.__name__], flag + bond + str(which_neighbour))[label]
    except KeyError:
        return 0


def d_me(N, l, m1, m2):
    """Computes rotational matrix elements according to
    A.V. Podolskiy and P. Vogl, Phys. Rev. B. 69, 233101 (2004)

    Parameters
    ----------
    N :
        directional cosine relative to z-axis
    l :
        orbital quantum number
    m1 :
        magnetic quantum number
    m2 :
        magnetic quantum number

    Returns
    -------
    type
        rotational matrix element

    """

    if N == -1:
       N += sys.float_info.epsilon


    prefactor = ((0.5 * (1 + N)) ** l) * (((1 - N) / (1 + N)) ** (m1 * 0.5 - m2 * 0.5)) * \
            math.sqrt(math.factorial(l + m2) * math.factorial(l - m2) *
                      math.factorial(l + m1) * math.factorial(l - m1))

    ans = 0

    for t in range(2 * l + 2):
        if l + m2 - t >= 0 and l - m1 - t >= 0 and t + m1 - m2 >= 0:
            if N == -1.0 and t == 0:
                ans += ((-1) ** t) / \
                       (math.factorial(l + m2 - t) * math.factorial(l - m1 - t) *
                        math.factorial(t) * math.factorial(t + m1 - m2))
            else:
                ans += ((-1) ** t) * (((1 - N) / (1 + N)) ** t) / \
                       (math.factorial(l + m2 - t) * math.factorial(l - m1 - t) *
                        math.factorial(t) * math.factorial(t + m1 - m2))

    return ans * prefactor


def tau(m):
    """Auxiliary function

    Parameters
    ----------
    m :
        

    Returns
    -------

    """
    if m < 0:
        return 0
    else:
        return 1


def a_coef(m, gamma):
    """

    Parameters
    ----------
    m :
        
    gamma :
        

    Returns
    -------

    """

    if m == 0:
        return 1.0 / math.sqrt(2)
    else:
        return ((-1) ** abs(m)) * \
               (tau(m) * math.cos(abs(m) * gamma) - tau(-m) * math.sin(abs(m) * gamma))


def b_coef(m, gamma):
    """

    Parameters
    ----------
    m :
        
    gamma :
        

    Returns
    -------

    """

    return ((-1) ** abs(m)) * \
           (tau(m) * math.sin(abs(m) * gamma) + tau(-m) * math.cos(abs(m) * gamma))


def s_me(N, l, m1, m2, gamma):
    """

    Parameters
    ----------
    N :
        
    l :
        
    m1 :
        
    m2 :
        
    gamma :
        

    Returns
    -------

    """

    return a_coef(m1, gamma) * \
           (((-1) ** abs(m2)) * d_me(N, l, abs(m1), abs(m2)) + d_me(N, l, abs(m1), -abs(m2)))


def t_me(N, l, m1, m2, gamma):
    """

    Parameters
    ----------
    N :
        
    l :
        
    m1 :
        
    m2 :
        
    gamma :
        

    Returns
    -------

    """

    if m1 == 0:
        return 0
    else:
        return b_coef(m1, gamma) * \
               (((-1) ** abs(m2)) * d_me(N, l, abs(m1), abs(m2)) - d_me(N, l, abs(m1), -abs(m2)))


def me(atom1, ll1, atom2, ll2, coords, which_neighbour=0, overlap=False):
    """Computes the non-diagonal matrix element of the tight-binding Hamiltonian -
    coupling between two sites, both are described by LCAO basis sets.
    This function is evoked in the member function _get_me() of the Hamiltonian object.

    Parameters
    ----------
    atom1 : tb.Orbitals
        basis set associated with the first site
    ll1 : int
        index specifying a particular orbital in the basis set for the first site
    atom2 : tb.Orbitals
        basis set associated with the first site
    ll2 : int
        index specifying a particular orbital in the basis set for the second site
    coords : array
        coordinates of radius vector pointing from one site to another
    which_neighbour : int
        Order of a nearest neighbour (first-, second-, third- etc) (Default value = 0)
    overlap : bool
            A flag indicating that the overlap matrix element has to be computed
    Returns
    -------

    
    """

    # determine type of bonds
    atoms = sorted([item.upper() for item in [atom1.title, atom2.title]])
    atoms = atoms[0] + '_' + atoms[1]

    # quantum numbers for the first atom
    n1 = atom1.orbitals[ll1]['n']
    l1 = atom1.orbitals[ll1]['l']
    m1 = atom1.orbitals[ll1]['m']
    s1 = atom1.orbitals[ll1]['s']

    # quantum numbers for the second atom
    n2 = atom2.orbitals[ll2]['n']
    l2 = atom2.orbitals[ll2]['l']
    m2 = atom2.orbitals[ll2]['m']
    s2 = atom2.orbitals[ll2]['s']

    if s1 == s2:

        L = coords[0]
        M = coords[1]
        N = coords[2]

        gamma = math.atan2(L, M)

        if l1 > l2:
            code = [n2, n1]
        elif l1 == l2:
            code = [min(n1, n2), max(n1, n2)]
        else:
            code = [n1, n2]

        for j, item in enumerate(code):
            if item == 0:
                code[j] = ""
            else:
                code[j] = str(item)

        l_min = min(l1, l2)
        l_max = max(l1, l2)

        prefactor = (-1) ** ((l1 - l2 + abs(l1 - l2)) * 0.5)
        ans = 2 * a_coef(m1, gamma) * a_coef(m2, gamma) * \
            d_me(N, l1, abs(m1), 0) * d_me(N, l2, abs(m2), 0) * \
            me_diatomic(atoms, code, l_min, l_max, 0, which_neighbour, overlap=overlap)

        for m in range(1, l_min+1):
            ans += (s_me(N, l1, m1, m, gamma) * s_me(N, l2, m2, m, gamma) +
                    t_me(N, l1, m1, m, gamma) * t_me(N, l2, m2, m, gamma)) * \
                   me_diatomic(atoms, code, l_min, l_max, m, which_neighbour, overlap=overlap)

        return prefactor * ans
    else:
        return 0


if __name__ == "__main__":

    x0 = np.array([0, 0, 0], dtype=float)
    x1 = np.array([0, 0, 1], dtype=float)

    coords = x0 - x1
    coords /= np.linalg.norm(coords)

    print(coords)

    print(d_me(-coords[2], 1, 1, 0))
    print(d_me(-coords[2], 1, 0, 1))
    print(d_me(-coords[2], 2, 1, 0))
    print(d_me(-coords[2], 2, 0, 1))
    print(d_me(-coords[2], 2, 2, 1))
    print(d_me(-coords[2], 2, 2, 1))
    print(d_me(-coords[2], 2, -1, 2))
    print(d_me(-coords[2], 2, 1, -2))
    print("-----------------------------")
    print(d_me(coords[2], 1, 1, 0))
    print(d_me(coords[2], 1, 0, 1))
    print(d_me(coords[2], 2, 1, 0))
    print(d_me(coords[2], 2, 0, 1))
    print(d_me(coords[2], 2, 2, 1))
    print(d_me(coords[2], 2, 2, 1))
    print(d_me(coords[2], 2, -1, 2))
    print(d_me(coords[2], 2, 1, -2))
    # print d_me(-coords[2], 1, -1, 0)
    # print d_me(-coords[2], 1, 0, -1)
    # print d_me(-coords[2], 1, -1, -1)
    # print d_me(-coords[2], 1, 1, 1)



from __future__ import division
import numpy as np
from scipy.spatial import cKDTree as KDTree


class Invdisttree:
    """Inverse-distance-weighted interpolation using KDTree:
    invdisttree = Invdisttree( X, z )  -- data points, values
    interpol = invdisttree( q, nnear=3, eps=0, p=1, weights=None, stat=0 )
    interpolates z from the 3 points nearest each query point q;
    For example, interpol[ a query point q ]
    finds the 3 data points nearest q, at distances d1 d2 d3
    and returns the IDW average of the values z1 z2 z3
        (z1/d1 + z2/d2 + z3/d3)
        / (1/d1 + 1/d2 + 1/d3)
        = .55 z1 + .27 z2 + .18 z3  for distances 1 2 3
    q may be one point, or a batch of points.
    eps: approximate nearest, dist <= (1 + eps) * true nearest
    p: use 1 / distance**p
    weights: optional multipliers for 1 / distance**p, of the same shape as q
    stat: accumulate wsum, wn for average weights
    How many nearest neighbors should one take ?
    a) start with 8 11 14 .. 28 in 2d 3d 4d .. 10d; see Wendel's formula
    b) make 3 runs with nnear= e.g. 6 8 10, and look at the results --
    |interpol 6 - interpol 8| etc., or |f - interpol*| if you have f(q).
    I find that runtimes don't increase much at all with nnear -- ymmv.
    p=1, p=2 ?
    p=2 weights nearer points more, farther points less.
    In 2d, the circles around query points have areas ~ distance**2,
    so p=2 is inverse-area weighting. For example,
        (z1/area1 + z2/area2 + z3/area3)
        / (1/area1 + 1/area2 + 1/area3)
        = .74 z1 + .18 z2 + .08 z3  for distances 1 2 3
    Similarly, in 3d, p=3 is inverse-volume weighting.
    Scaling:
    if different X coordinates measure different things, Euclidean distance
    can be way off.  For example, if X0 is in the range 0 to 1
    but X1 0 to 1000, the X1 distances will swamp X0;
    rescale the data, i.e. make X0.std() ~= X1.std() .
    A nice property of IDW is that it's scale-free around query points:
    if I have values z1 z2 z3 from 3 points at distances d1 d2 d3,
    the IDW average
    (z1/d1 + z2/d2 + z3/d3)
    / (1/d1 + 1/d2 + 1/d3)
    is the same for distances 1 2 3, or 10 20 30 -- only the ratios matter.
    In contrast, the commonly-used Gaussian kernel exp( - (distance/h)**2 )
    is exceedingly sensitive to distance and to h.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__( self, X, z, leafsize=10, stat=0 ):
        assert len(X) == len(z), "len(X) %d != len(z) %d" % (len(X), len(z))
        self.tree = KDTree( X, leafsize=leafsize )  # build the tree
        self.z = z
        self.stat = stat
        self.wn = 0
        self.wsum = None;

    def __call__( self, q, nnear=6, eps=0, p=1, weights=None ):
            # nnear nearest neighbours of each query point --
        q = np.asarray(q)
        qdim = q.ndim
        if qdim == 1:
            q = np.array([q])
        if self.wsum is None:
            self.wsum = np.zeros(nnear)

        self.distances, self.ix = self.tree.query( q, k=nnear, eps=eps )
        interpol = np.zeros( (len(self.distances),) + np.shape(self.z[0]) )
        jinterpol = 0
        for dist, ix in zip( self.distances, self.ix ):
            if nnear == 1:
                wz = self.z[ix]
            elif dist[0] < 1e-10:
                wz = self.z[ix[0]]
            else:  # weight z s by 1/dist --
                w = 1 / dist**p
                if weights is not None:
                    w *= weights[ix]  # >= 0
                w /= np.sum(w)
                wz = np.dot( w, self.z[ix] )
                if self.stat:
                    self.wn += 1
                    self.wsum += w
            interpol[jinterpol] = wz
            jinterpol += 1
        return interpol if qdim > 1  else interpol[0]


import logging
import numpy as np
from scipy.linalg import block_diag
import matplotlib.pyplot as plt
from nanonet.tb.hamiltonian import Hamiltonian
import nanonet.tb as tb
import nanonet.negf as negf
from nanonet.tb.aux_functions import fd
from nanonet.tb.diatomic_matrix_element import me
from nanonet.negf.recursive_greens_functions import recursive_gf
from nanonet.negf.hamiltonian_chain import HamiltonianChain
from ase.visualize import view


class Lead(Hamiltonian):

    def __init__(self, **kwargs):

        logging.info("Making leads.... ")

        super(Lead, self).__init__(**kwargs)
        self.initialize()

        period = kwargs.get('period', [[0, 0, 1.0]])
        if isinstance(period, float) or isinstance(period, int):
            self.set_periodic_bc([[0, 0, period]])
        else:
            self.set_periodic_bc(period)

        self._h_l, self._h_0, self._h_r = self.get_hamiltonians()

        logging.info("Done!")

    @property
    def h_0(self):
        return self._h_0

    @property
    def h_l(self):
        return self._h_l

    @property
    def h_r(self):
        return self._h_r

    def get_self_energy(self, energy, iterate=True, damp=0.00005j):

        flag = False

        if isinstance(energy, (int, float)):
            energy = [energy]
            flag = True

        sgf_l = []
        sgf_r = []

        for E in energy:
            L, R = negf.surface_greens_function(E, self._h_l, self._h_0, self._h_r, iterate=iterate, damp=damp)
            sgf_l.append(L)
            sgf_r.append(R)

        sgf_l = np.array(sgf_l)
        sgf_r = np.array(sgf_r)

        if flag:
            sgf_l[0] = sgf_l
            sgf_r[0] = sgf_r

        return sgf_l, sgf_r


class TwoTerminalDevice(object):

    def __init__(self, ll, dev, rl):

        self._device = dev
        self._left_lead = ll
        self._right_lead = rl
        self._field = None
        self._h_l = None
        self._h_r = None
        self._h_0 = None

        self._initialize()

    def _initialize(self):
        """Compute matrix elements of the Hamiltonian.

        Returns
        -------
        type Hamiltonian
            Returns the instance of the class Hamiltonian
        """

        self._h_0 = [self._left_lead.h_0, self._device.h_matrix, self._right_lead.h_0]

        self._h_l = np.zeros((self._left_lead.basis_size, self._device.basis_size), dtype=complex)
        self._h_r = np.zeros((self._device.basis_size, self._right_lead.basis_size), dtype=complex)

        # if self.compute_overlap:
        #     self.ov_matrix = np.zeros((self.basis_size, self.basis_size), dtype=complex)
        #     self.ov_matrix_bc_add = np.zeros((self.basis_size, self.basis_size), dtype=complex)

        # loop over all nodes
        for j1 in range(self._left_lead.num_of_nodes):
            # find neighbours for each node
            list_of_neighbours = self._device.get_neighbours(self._left_lead._coords[j1])[1:]
            print(list_of_neighbours)

            for j2 in list_of_neighbours:
                # nearest neighbours interaction
                for l1 in range(self._left_lead.orbitals_dict[list(self._left_lead.atom_list.keys())[j1]].num_of_orbitals):
                    for l2 in range(self._device.orbitals_dict[list(self._device.atom_list.keys())[j2]].num_of_orbitals):
                        ind1 = self._left_lead.qn2ind([('atoms', j1), ('l', l1)], )
                        ind2 = self._device.qn2ind([('atoms', j2), ('l', l2)], )

                        atom_kind1 = self._left_lead._ind2atom(j1)
                        atom_kind2 = self._device._ind2atom(j2)

                        coords1 = np.array(list(self._left_lead.atom_list.values())[j1], dtype=float) - \
                                  np.array(list(self._device.atom_list.values())[j2], dtype=float)

                        norm = np.linalg.norm(coords1)

                        if self._device.int_radial_dependence is None:
                            which_neighbour = ""
                        else:
                            which_neighbour = self._device.int_radial_dependence(norm)

                        if self._device.radial_dependence is None:
                            factor = 1.0
                        else:
                            factor = self._device.radial_dependence(norm)

                        # compute directional cosines
                        if self._device.compute_angular:
                            coords1 /= norm
                        else:
                            coords1 = np.array([1.0, 0.0, 0.0])

                        self._h_l[ind1, ind2] = me(atom_kind1, l1, atom_kind2, l2, coords1, which_neighbour, overlap=False) * factor

        # loop over all nodes
        for j1 in range(self._right_lead.num_of_nodes):
            # find neighbours for each node
            list_of_neighbours = self._device.get_neighbours(self._right_lead._coords[j1])[1:]
            print(list_of_neighbours)

            for j2 in list_of_neighbours:
                # nearest neighbours interaction
                for l1 in range(self._right_lead.orbitals_dict[
                                    list(self._right_lead.atom_list.keys())[j1]].num_of_orbitals):
                    for l2 in range(self._device.orbitals_dict[
                                        list(self._device.atom_list.keys())[j2]].num_of_orbitals):
                        ind1 = self._right_lead.qn2ind([('atoms', j1), ('l', l1)], )
                        ind2 = self._device.qn2ind([('atoms', j2), ('l', l2)], )

                        atom_kind1 = self._right_lead._ind2atom(j1)
                        atom_kind2 = self._device._ind2atom(j2)

                        coords1 = np.array(list(self._right_lead.atom_list.values())[j1], dtype=float) - \
                                  np.array(list(self._device.atom_list.values())[j2], dtype=float)

                        norm = np.linalg.norm(coords1)

                        if self._device.int_radial_dependence is None:
                            which_neighbour = ""
                        else:
                            which_neighbour = self._device.int_radial_dependence(norm)

                        if self._device.radial_dependence is None:
                            factor = 1.0
                        else:
                            factor = self._device.radial_dependence(norm)

                        # compute directional cosines
                        if self._device.compute_angular:
                            coords1 /= norm
                        else:
                            coords1 = np.array([1.0, 0.0, 0.0])

                        self._h_r[ind2, ind1] = me(atom_kind1, l1, atom_kind2, l2, coords1,
                                                   which_neighbour, overlap=False) * factor

                        # if self.compute_overlap:
                        #     self.ov_matrix[ind1, ind2] = self._get_me(j1, j2, l1, l2, overlap=True)

        self._h_l = [self._h_l, self._h_r]
        self._h_r = [self._h_l[0].conj().T, self._h_r.conj().T]

        # self._h_r = [self._h_r, self._h_l.conj().T]
        # self._h_l = [self._h_r[0].conj().T, self._h_l]

    def sgf(self, energy, ef1, ef2, tempr, sgf_l, sgf_r):
        """ """

        sgf = [None for _ in range(len(self._h_0))]

        for jjj in range(len(self._h_0)):
            if jjj == 0:
                sgf[jjj] = -2.0 * np.imag(sgf_l) * fd(energy, ef1, tempr)
            elif jjj == len(self._h_0) - 1:
                sgf[jjj] = -2.0 * np.imag(sgf_r) * fd(energy, ef2, tempr)
            else:
                sgf[jjj] = np.zeros(self._h_0[jjj].shape)

        return sgf

    def comp_dos_tr(self, energy):

        dos = np.zeros(len(energy))
        tr = np.zeros(len(energy))

        for j, en in enumerate(energy):

            L, _ = self._left_lead.get_self_energy(en)
            _, R = self._right_lead.get_self_energy(en)

            L = L[0]
            R = R[0]

            self._h_0[-1] = self._h_0[-1] + R
            self._h_0[0] = self._h_0[0] + L

            g_trans, grd, grl, gru, gr_left = recursive_gf(en,
                                                           self._h_r,
                                                           self._h_0,
                                                           self._h_l,
                                                           damp=0.00005j)

            self._h_0[-1] = self._h_0[-1] - R
            self._h_0[0] = self._h_0[0] - L

            gamma_l = 1j * (L - L.conj().T)
            gamma_r = 1j * (R - R.conj().T)

            tr[j] = np.real(np.trace(gamma_r.dot(g_trans.conj().T).dot(gamma_l).dot(g_trans)))

            for jj in range(len(self._h_0)):
                dos[j] = dos[j] + np.real(np.trace(1j * (grd[jj] - grd[jj].conj().T))) / len(self._h_0)

        return dos, tr

    def comp_dos_tr_dense(self, energy, ef1, ef2, tempr):

        dos = np.zeros(len(energy))
        tr = np.zeros(len(energy))
        dens = np.zeros((len(energy), len(self._h_0)))
        dens = np.zeros((len(energy), len(self._h_0[0]) + len(self._h_0[1]) + len(self._h_0[2])))

        for j, en in enumerate(energy):

            L, _ = self._left_lead.get_self_energy(en)
            _, R = self._right_lead.get_self_energy(en)

            L = L[0]
            R = R[0]

            self._h_0[-1] = self._h_0[-1] + R
            self._h_0[0] = self._h_0[0] + L

            g_trans, grd, grl, gru, gr_left, gnd, gnl, gnu, gn_left = recursive_gf(en,
                                                                                   self._h_r,
                                                                                   self._h_0,
                                                                                   self._h_l,
                                                                                   s_in=self.sgf(en, ef1, ef2, tempr, L, R),
                                                                                   damp=0.00005j)

            self._h_0[-1] = self._h_0[-1] - R
            self._h_0[0] = self._h_0[0] - L

            gamma_l = 1j * (L - L.conj().T)
            gamma_r = 1j * (R - R.conj().T)

            tr[j] = np.real(np.trace(gamma_r.dot(g_trans.conj().T).dot(gamma_l).dot(g_trans)))

            for jj in range(len(self._h_0)):
                dos[j] = dos[j] + np.real(np.trace(1j * (grd[jj] - grd[jj].conj().T))) / len(self._h_0)

            dens[j, :] = 2 * np.hstack((np.diag(gnd[0]), np.diag(gnd[1]), np.diag(gnd[2])))

        return dos, tr, dens

    # @property
    # def field(self):
    #     return self._field
    #
    # @property
    # def device(self):
    #     return self._device
    #
    # @property
    # def left_lead(self):
    #     return self._left_lead
    #
    # @property
    # def right_lead(self):
    #     return self._right_lead
    #
    # @field.setter
    # def field(self, value):
    #     pass
    #
    # @device.setter
    # def device(self, value):
    #     self._device = value
    #
    # @left_lead.setter
    # def left_lead(self, value):
    #     pass
    #
    # @right_lead.setter
    # def right_lead(self, value):
    #     pass
    #
    # @field.deleter
    # def field(self):
    #     pass
    #
    # @left_lead.deleter
    # def left_lead(self):
    #     pass
    #
    # @right_lead.deleter
    # def right_lead(self):
    #     pass
    #
    # @device.deleter
    # def device(self):
    #     pass

    def get_matrix(self):
        """ """

        ans = block_diag(*tuple(self._h_0))

        y0 = 0
        x0 = self._h_r[0].shape[1]

        for item in self._h_r:
            ans[x0:x0 + item.shape[0], y0:y0 + item.shape[1]] = item
            x0 = x0 + item.shape[0]
            y0 = y0 + item.shape[1]

        x0 = 0
        y0 = self._h_l[0].shape[0]

        for item in self._h_l:
            ans[x0:x0 + item.shape[0], y0:y0 + item.shape[1]] = item
            x0 = x0 + item.shape[0]
            y0 = y0 + item.shape[1]

        return ans

    def visualize(self):
        """ """

        import matplotlib.pyplot as plt
        from matplotlib import cm

        vals = np.zeros(len(self.fields[0])*len(self.fields[0][0]))

        for field in self.fields:
            vals += np.hstack(tuple([item for item in field]))

        vals /= np.max(vals)

        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        norm = cm.colors.Normalize(vmax=np.abs(np.max(vals)), vmin=-np.abs(np.min(vals)))

        ax.scatter(self.coords[:, 0], self.coords[:, 1], self.coords[:, 2],
                   c=vals,
                   marker='o',
                   norm=norm,
                   cmap=cm.get_cmap(cm.coolwarm))

        plt.show()

    @property
    def coords(self):
        """ """
        if self.elem_length is not None:
            coords = [self._coords - jjj * self.elem_length for jjj in range(self.left_translations, 0, -1)] + \
                     [self._coords] + \
                     [self._coords + jjj * self.elem_length for jjj in range(1, self.right_translations + 1)]

            return np.concatenate(coords)
        else:
            return self._coords


def main():

    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)
    b = tb.Orbitals('B')
    b.add_orbital('s', -0.5)
    c = tb.Orbitals('C')
    c.add_orbital('s', -0.3)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5},
                     PARAMS_B_B={'ss_sigma': -0.5},
                     PARAMS_A_B={'ss_sigma': -0.5},
                     PARAMS_B_C={'ss_sigma': -0.5},
                     PARAMS_A_C={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    5.0    4.0
                          B1    0.0    5.0    5.0
                          A2    0.0    6.0    4.0
                          B2    0.0    6.0    5.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    0.0
                          B1    0.0    0.0    1.0
                          A2    0.0    1.0    0.0
                          B2    0.0    1.0    1.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""14
                                   H cell
                                   A1    0.0    0.0    2.0
                                   B1    0.0    0.0    3.0
                                   A2    0.0    1.0    2.0
                                   B2    0.0    1.0    3.0
                                   A3    0.0    2.0    2.0
                                   B3    0.0    2.0    3.0
                                   A4    0.0    3.0    2.0
                                   B4    0.0    3.0    3.0
                                   A5    0.0    4.0    2.0
                                   B5    0.0    4.0    3.0
                                   A6    0.0    5.0    2.0
                                   B6    0.0    5.0    3.0
                                   A7    0.0    6.0    2.0
                                   B7    0.0    6.0    3.0
                                """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    plt.figure(1)
    plt.plot(np.log(tr))
    plt.figure(2)
    plt.plot(np.log(dos))
    plt.show()


def main0():

    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)
    b = tb.Orbitals('B')
    b.add_orbital('s', -0.5)
    c = tb.Orbitals('C')
    c.add_orbital('s', -0.3)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5},
                     PARAMS_B_B={'ss_sigma': -0.5},
                     PARAMS_A_B={'ss_sigma': -0.5},
                     PARAMS_B_C={'ss_sigma': -0.5},
                     PARAMS_A_C={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    5.0    -6.0
                          B1    0.0    5.0    -5.0
                          A2    0.0    6.0    -6.0
                          B2    0.0    6.0    -5.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    -10.0
                          B1    0.0    0.0    -9.0
                          A2    0.0    1.0    -10.0
                          B2    0.0    1.0    -9.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""14
                                   H cell
                                   A1    0.0    0.0    -8.0
                                   B1    0.0    0.0    -7.0
                                   A2    0.0    1.0    -8.0
                                   B2    0.0    1.0    -7.0
                                   A3    0.0    2.0    -8.0
                                   B3    0.0    2.0    -7.0
                                   A4    0.0    3.0    -8.0
                                   B4    0.0    3.0    -7.0
                                   A5    0.0    4.0    -8.0
                                   B5    0.0    4.0    -7.0
                                   A6    0.0    5.0    -8.0
                                   B6    0.0    5.0    -7.0
                                   A7    0.0    6.0    -8.0
                                   B7    0.0    6.0    -7.0
                                """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main1():
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)
    b = tb.Orbitals('B')
    b.add_orbital('s', -0.5)
    c = tb.Orbitals('C')
    c.add_orbital('s', -0.3)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5},
                     PARAMS_B_B={'ss_sigma': -0.5},
                     PARAMS_A_B={'ss_sigma': -0.5},
                     PARAMS_B_C={'ss_sigma': -0.5},
                     PARAMS_A_C={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    4.0
                          B1    0.0    0.0    5.0
                          A2    0.0    1.0    4.0
                          B2    0.0    1.0    5.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    0.0
                          B1    0.0    0.0    1.0
                          A2    0.0    1.0    0.0
                          B2    0.0    1.0    1.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""4
                                   H cell
                                   A1    0.0    0.0    2.0
                                   B1    0.0    0.0    3.0
                                   A2    0.0    1.0    2.0
                                   B2    0.0    1.0    3.0
                                """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main1_1():
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)
    b = tb.Orbitals('B')
    b.add_orbital('s', -0.5)
    c = tb.Orbitals('C')
    c.add_orbital('s', -0.3)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5},
                     PARAMS_B_B={'ss_sigma': -0.5},
                     PARAMS_A_B={'ss_sigma': -0.5},
                     PARAMS_B_C={'ss_sigma': -0.5},
                     PARAMS_A_C={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    4.0
                          B1    0.0    0.0    5.0
                          A2    0.0    1.0    4.0
                          B2    0.0    1.0    5.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 2.0]],
                   xyz="""4
                          H cell
                          A1    0.0    0.0    0.0
                          B1    0.0    0.0    1.0
                          A2    0.0    1.0    0.0
                          B2    0.0    1.0    1.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""4
                                   H cell
                                   A1    0.0    0.0    2.0
                                   B1    0.0    0.0    3.0
                                   A2    0.0    1.0    2.0
                                   B2    0.0    1.0    3.0
                                """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main2():
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)

    tb.Orbitals.orbital_sets = {'A': a}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 1.0]],
                   xyz="""1
                          A cell
                          A1    0.0    4.0    2.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 1.0]],
                   xyz="""1
                          A cell
                          A1    0.0    10.0    0.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""9
                                   H cell
                                   A4    0.0    3.0    1.0
                                   A5    0.0    4.0    1.0
                                   A6    0.0    5.0    1.0
                                   A7    0.0    6.0    1.0
                                   A8    0.0    7.0    1.0
                                   A9    0.0    8.0    1.0
                                   A10    0.0    9.0    1.0
                                   A11    0.0    10.0    1.0
                                   A12    0.0    11.0    1.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main3():
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.7)

    b = tb.Orbitals('B')
    b.add_orbital('s', -0.3)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.5}, PARAMS_A_B={'ss_sigma': -0.5}, PARAMS_B_B={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0, 1.0]],
                   xyz="""1
                          A cell
                          A1    0.0    4.0    2.0
                       """)

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0, 1.0]],
                   xyz="""1
                          A cell
                          A1    0.0    10.0    0.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""9
                                   H cell
                                   A4    0.0    3.0    1.0
                                   B5    0.0    4.0    1.0
                                   A6    0.0    5.0    1.0
                                   B7    0.0    6.0    1.0
                                   A8    0.0    7.0    1.0
                                   B9    0.0    8.0    1.0
                                   A10    0.0    9.0    1.0
                                   B11    0.0    10.0    1.0
                                   A12    0.0    11.0    1.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-3.0, 1.5, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main4(energy):
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.2)

    b = tb.Orbitals('B')
    b.add_orbital('s', 0.2)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b}

    tb.set_tb_params(PARAMS_A_B={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""2
                          A cell
                          A    0.0    1.0    1.0
                          B    0.0    1.0    2.0
                       """)

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""2
                          A cell
                          B    0.0    1.0    12.0
                          A    0.0    1.0    13.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""9
                                   H cell
                                   A    0.0    1.0    3.0
                                   B    0.0    1.0    4.0
                                   A    0.0    1.0    5.0
                                   B    0.0    1.0    6.0
                                   A    0.0    1.0    7.0
                                   B    0.0    1.0    8.0
                                   A    0.0    1.0    9.0
                                   B    0.0    1.0    10.0
                                   A    0.0    1.0    11.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main5(energy):
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', 0)

    tb.Orbitals.orbital_sets = {'A': a}

    tb.set_tb_params(PARAMS_A_A={'ss_sigma': -0.7})

    # --------------------- make leads ---------------------

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          A    0.0    1.0    2.0
                       """)

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          A    0.0    1.0    12.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""9
                                   H cell
                                   A    0.0    1.0    3.0
                                   A    0.0    1.0    4.0
                                   A    0.0    1.0    5.0
                                   A    0.0    1.0    6.0
                                   A    0.0    1.0    7.0
                                   A    0.0    1.0    8.0
                                   A    0.0    1.0    9.0
                                   A    0.0    1.0    10.0
                                   A    0.0    1.0    11.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main6():
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.2)

    b = tb.Orbitals('B')
    b.add_orbital('s', 0.2)

    c = tb.Orbitals('C')
    c.add_orbital('s', 0)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_C={'ss_sigma': -0.61}, PARAMS_C_C={'ss_sigma': -0.7}, PARAMS_A_B={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          C    0.0    1.0    2.0
                       """)

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          C    0.0    1.0    20.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""17
                                   H cell
                                   A    0.0    1.0    3.0
                                   B    0.0    1.0    4.0
                                   A    0.0    1.0    5.0
                                   B    0.0    1.0    6.0
                                   A    0.0    1.0    7.0
                                   B    0.0    1.0    8.0
                                   A    0.0    1.0    9.0
                                   B    0.0    1.0    10.0
                                   A    0.0    1.0    11.0
                                   B    0.0    1.0    12.0
                                   A    0.0    1.0    13.0
                                   B    0.0    1.0    14.0
                                   A    0.0    1.0    15.0
                                   B    0.0    1.0    16.0
                                   A    0.0    1.0    17.0
                                   B    0.0    1.0    18.0
                                   A    0.0    1.0    19.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    energy = np.linspace(-2.0, 2.0, 1700)
    dos, tr = a.comp_dos_tr(energy)

    return dos, tr


def main7(energy):
    # --------------------- make orbitals ---------------------

    a = tb.Orbitals('A')
    a.add_orbital('s', -0.2)

    b = tb.Orbitals('B')
    b.add_orbital('s', 0.2)

    c = tb.Orbitals('C')
    c.add_orbital('s', 0)

    tb.Orbitals.orbital_sets = {'A': a, 'B': b, 'C': c}

    tb.set_tb_params(PARAMS_A_C={'ss_sigma': -0.61}, PARAMS_C_C={'ss_sigma': -0.7}, PARAMS_A_B={'ss_sigma': -0.5})

    # --------------------- make leads ---------------------

    leads_l = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          C    0.0    1.0    2.0
                       """)

    leads_r = Lead(nn_distance=1.1,
                   period=[[0, 0.0, 1.0]],
                   xyz="""1
                          A cell
                          C    0.0    1.0    20.0
                       """)

    # -------------- make two terminal device --------------

    device = tb.Hamiltonian(nn_distance=1.1,
                            xyz="""17
                                   H cell
                                   A    0.0    1.0    3.0
                                   B    0.0    1.0    4.0
                                   A    0.0    1.0    5.0
                                   B    0.0    1.0    6.0
                                   A    0.0    1.0    7.0
                                   B    0.0    1.0    8.0
                                   A    0.0    1.0    9.0
                                   B    0.0    1.0    10.0
                                   A    0.0    1.0    11.0
                                   B    0.0    1.0    12.0
                                   A    0.0    1.0    13.0
                                   B    0.0    1.0    14.0
                                   A    0.0    1.0    15.0
                                   B    0.0    1.0    16.0
                                   A    0.0    1.0    17.0
                                   B    0.0    1.0    18.0
                                   A    0.0    1.0    19.0
                                """)

    # device = tb.Hamiltonian(nn_distance=1.1,
    #                         xyz="""4
    #                                H cell
    #                                A1    0.0    0.0    2.0
    #                                B1    0.0    0.0    3.0
    #                                A2    0.0    1.0    2.0
    #                                B2    0.0    1.0    3.0
    #                             """)

    device.initialize()

    a = TwoTerminalDevice(leads_l, device, leads_r)
    print(a._h_l)
    print(a._h_0)
    print(a._h_r)
    plt.imshow(np.real(a.get_matrix()))
    plt.show()
    ef1 = 0
    ef2 = 0
    tempr = 300
    dos, tr, dense = a.comp_dos_tr_dense(energy, ef1, ef2, tempr)

    return dos, tr, dense


if __name__ == '__main__':

    # main2()
    # dos, tr = main0()
    # dos1, tr1 = main1()
    #
    # plt.figure(1)
    # plt.plot(np.log(tr))
    # plt.plot(np.log(tr1))
    # plt.figure(2)
    # plt.plot(np.log(dos))
    # plt.plot(np.log(dos1))
    # plt.show()

    # dos, tr = main6()
    #
    # plt.figure(1)
    # plt.plot(tr)
    # # plt.yscale('log')
    #
    # plt.figure(2)
    # plt.plot(dos)
    # # plt.yscale('log')
    # plt.show()

    energy = np.linspace(-2.0, 2.0, 1700)

    dos1, tr1 = main4(energy)
    dos2, tr2 = main5(energy)

    plt.figure(1)
    plt.plot(tr1)
    plt.plot(tr2)
    plt.yscale('log')

    plt.figure(2)
    plt.plot(dos1)
    plt.plot(dos2)
    plt.yscale('log')
    plt.show()

    dos3, tr3, dens = main7(energy)

    plt.figure(3)
    plt.plot(tr3)
    plt.figure(4)
    plt.plot(dos3)
    plt.show()



"""
The module contains all necessary classes needed to compute the Hamiltonian matrix
"""
from __future__ import print_function, division
from __future__ import absolute_import
import matplotlib.pyplot as plt
import numpy as np
import scipy.sparse.linalg as splin
import scipy.sparse as sp
from nanonet.tb.orbitals import Orbitals
from nanonet.tb.hamiltonian import Hamiltonian


class HamiltonianSp(Hamiltonian):
    """Class defines a Hamiltonian matrix as well as a set of member-functions
    allowing to build, diagonalize and visualize the matrix.

    Parameters
    ----------

    Returns
    -------

    """

    def __init__(self, **kwargs):

        super(HamiltonianSp, self).__init__(**kwargs)

        sigma = kwargs.get('sigma', 1.1)
        num_eigs = kwargs.get('num_eigs', 14)

        self.sigma = sigma
        self.num_eigs = num_eigs

    def initialize(self):
        """The function computes matrix elements of the Hamiltonian."""

        # initialize Hamiltonian matrices
        self.h_matrix = sp.lil_matrix((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_add = sp.lil_matrix((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = np.ones((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = sp.lil_matrix(self.h_matrix_bc_factor)

        # loop over all nodes
        for j1 in range(self.num_of_nodes):

            # find neighbours for each node
            list_of_neighbours = self.get_neighbours(j1)

            for j2 in list_of_neighbours:
                # on site interactions
                if j1 == j2:
                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        ind1 = self.qn2ind([('atoms', j1), ('l', l1)], )
                        self.h_matrix[ind1, ind1] = self._get_me(j1, j2, l1, l1)

                # nearest neighbours interaction
                else:
                    for l1 in range(self.orbitals_dict[list(self.atom_list.keys())[j1]].num_of_orbitals):
                        for l2 in range(self.orbitals_dict[list(self.atom_list.keys())[j2]].num_of_orbitals):

                            ind1 = self.qn2ind([('atoms', j1), ('l', l1)], )
                            ind2 = self.qn2ind([('atoms', j2), ('l', l2)], )

                            self.h_matrix[ind1, ind2] = self._get_me(j1, j2, l1, l2)

        return self

    def diagonalize(self):
        """Diagonalize the Hamiltonian matrix for the finite isolated system
        :return:

        Parameters
        ----------

        Returns
        -------

        """

        vals, vects = splin.eigsh(self.h_matrix, k=self.num_eigs, sigma=self.sigma)
        vals = np.real(vals)
        ind = np.argsort(vals)
        return vals[ind], vects[:, ind]

    def diagonalize_periodic_bc(self, k_vector):
        """Diagonalize the Hamiltonian matrix with the periodic boundary conditions
        for a certain value of the wave vector k_vector

        Parameters
        ----------
        k_vector :
            wave vector

        Returns
        -------

        """

        k_vector = list(k_vector)

        # reset previous wave vector if any
        if k_vector != self.k_vector:

            self._reset_periodic_bc()
            self.k_vector = k_vector
            self._compute_h_matrix_bc_factor()
            self._compute_h_matrix_bc_add()

        try:
            vals, vects = splin.eigsh(self.h_matrix_bc_factor.multiply(self.h_matrix) + self.h_matrix_bc_add,
                                      k=self.num_eigs, sigma=self.sigma)
        except TypeError:
            mat = self.h_matrix_bc_factor.multiply(self.h_matrix) + self.h_matrix_bc_add
            vals, vects = np.linalg.eigh(mat.todense())

        vals = np.real(vals)
        ind = np.argsort(vals)

        return vals[ind], vects[:, ind]

    def _reset_periodic_bc(self):
        """Resets the matrices determining periodic boundary conditions to their default state
        :return:

        Parameters
        ----------

        Returns
        -------

        """

        self.h_matrix_bc_add = sp.lil_matrix((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = np.ones((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_bc_factor = sp.lil_matrix(self.h_matrix_bc_factor)
        self.k_vector = None

    def get_hamiltonians(self):
        """ """

        self.k_vector = [0.0, 0.0, 0.0]

        self.h_matrix_left_lead = sp.lil_matrix((self.basis_size, self.basis_size), dtype=complex)
        self.h_matrix_right_lead = sp.lil_matrix((self.basis_size, self.basis_size), dtype=complex)

        self._compute_h_matrix_bc_add(split_the_leads=True)
        self.k_vector = None

        return self.h_matrix_left_lead.T, self.h_matrix, self.h_matrix_right_lead.T


def main():
    """ """

    a_si = 5.50
    PRIMITIVE_CELL = [[0, 0, a_si]]
    Orbitals.orbital_sets = {'Si': 'SiliconSP3D5S', 'H': 'HydrogenS'}

    h = HamiltonianSp(xyz='/home/mk/TB_project/input_samples/SiNW.xyz')
    h.initialize()
    h.set_periodic_bc(PRIMITIVE_CELL)

    num_points = 30
    kk = np.linspace(0, np.PI / a_si, num_points, endpoint=True)
    band_sructure = []

    for jj in range(num_points):
        vals, _ = h.diagonalize_periodic_bc([0.0, 0.0, kk[jj]])
        band_sructure.append(vals)

    band_sructure = np.array(band_sructure)

    vb = np.sort(np.real(band_sructure))
    cb = vb.copy()
    vb[vb > 0] = np.NaN
    cb[cb < 0] = np.NaN

    vb = -np.sort(-vb)
    cb = np.sort(cb)

    fig, ax = plt.subplots(1, 2)
    ax[0].set_ylim(-1.0, -0.3)
    ax[0].plot(kk, vb)
    ax[0].set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
    ax[0].set_ylabel(r'Energy (eV)')
    ax[0].set_title('Valence band')

    ax[1].set_ylim(2.0, 2.7)
    ax[1].plot(kk, cb)
    ax[1].set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
    ax[1].set_ylabel(r'Energy (eV)')
    ax[1].set_title('Conduction band')
    fig.tight_layout()
    plt.savefig('test.pdf', format='pdf')
    plt.show()


if __name__ == '__main__':

    main()



"""This module contains three sorting function: lexicographic sort of atomic coordinates,
sort that uses projections on a vector pointing from one electrode to another as the sorting keys
and sort that uses a potential function over atomic coordinates as the sorting keys.
A user can define his own sorting procedure - the user-defined sorting function should contain
`**kwargs` in the list of arguments and it can uses in its body one of the arguments with following name convention:
`coords` is the list of atomic coordinates,
`left_lead` is the list of the indices of the atoms contacting the left lead,
`right_lead` is the list of the indices of the atoms contacting the right lead, and
`mat` is the adjacency matrix of the tight-binding model.
All functions return the list of sorted atomic indices.
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.sparse.linalg import lgmres


def sort_lexico(coords=None, **kwargs):
    """Lexicographic sort

    Parameters
    ----------
    coords : array
        list of atomic coordinates (Default value = None)
    **kwargs :
        

    Returns
    -------

    
    """
    return np.lexsort((coords[:, 0], coords[:, 1], coords[:, 2]))


def sort_projection(coords=None, left_lead=None, right_lead=None, **kwargs):
    """Sorting procedure that uses projections on a vector pointing from one electrode to another as the sorting keys.

    Parameters
    ----------
    coords : array
        list of atomic coordinates (Default value = None)
    left_lead : array
        list of the atom indices contacting the left lead (Default value = None)
    right_lead : array
        list of the atom indices contacting the right lead (Default value = None)
    **kwargs :
        

    Returns
    -------

    
    """
    vec = np.mean(coords[left_lead], axis=0) - np.mean(coords[right_lead], axis=0)
    keys = np.dot(coords, vec) / np.linalg.norm(vec)

    return np.argsort(keys, kind='mergesort')


def sort_capacitance(coords, mat, left_lead, right_lead, **kwargs):
    """Sorting procedure that uses a potential function defined over atomic coordinates as the sorting keys.

    Parameters
    ----------
    coords : array
        list of atomic coordinates
    mat : 2D array
        adjacency matrix of the tight-binding model
    left_lead : array
        list of the atom indices contacting the left lead
    right_lead : array
        list of the atom indices contacting the right lead
    **kwargs :
        

    Returns
    -------

    
    """

    charge = np.zeros(coords.shape[0], dtype=complex)
    charge[left_lead] = 1e3
    charge[right_lead] = -1e3

    x = coords[:, 1].T
    y = coords[:, 0].T

    mat = (mat != 0.0).astype(float)
    mat = 10 * (mat - np.diag(np.diag(mat)))
    mat = mat - np.diag(np.sum(mat, axis=1)) + 0.001 * np.identity(mat.shape[0])

    col, info = lgmres(mat, charge.T, x0=1.0 / np.diag(mat), tol=1e-5, maxiter=15)
    col = col / np.max(col)

    indices = np.argsort(col, kind='heapsort')

    mat = mat[indices, :]
    mat = mat[:, indices]

    plt.scatter(x, y, c=col, cmap=plt.cm.get_cmap('seismic'), s=50, marker="o", edgecolors="k")
    plt.colorbar()
    plt.axis('off')
    plt.show()

    return indices



#!/usr/bin/env python
from __future__ import print_function
import argparse
import pickle
import matplotlib.pyplot as plt
import numpy as np
import nanonet.tb as tb


def preprocess_data(param_file, k_points_file, xyz, code_name):
    """

    Parameters
    ----------
    param_file :
        
    k_points_file :
        
    xyz :
        
    code_name :
        

    Returns
    -------

    """
    params = tb.yaml_parser(param_file)    # parse parameter file

    if k_points_file is None:   # default k-points
        if len(params['primitive_cell']) == 3:
            sym_points = ['L', 'GAMMA', 'X', 'W', 'K', 'L', 'W', 'X', 'K', 'GAMMA']
            num_points = [15, 20, 15, 10, 15, 15, 15, 15, 20]
            wave_vector = tb.get_k_coords(sym_points, num_points, 'Si')
        else:
            wave_vector = [[0.0, 0.0, 0.0]]
    else:   # read k-points file if its path is provided from the command line arguments
        wave_vector = np.loadtxt(k_points_file)

    # read xyz file if its path is provided from the command line arguments
    if isinstance(xyz, str):
        with open(xyz, 'r') as myfile:
            params['xyz'] = myfile.read()

    if not isinstance(code_name, str):
        code_name = 'band_structure'

    return params, wave_vector, code_name


def postprocess_data(kk, band_structure, show, save, code_name):
    """

    Parameters
    ----------
    kk :
        
    band_structure :
        
    show :
        
    save :
        
    code_name :
        

    Returns
    -------

    """

    flag = True  # 1D system

    if not isinstance(band_structure, np.ndarray):
        ids = [band_structure[item]['id'] for item in range(len(band_structure))]
        band_structure = [x['eigenvalues'] for _, x in sorted(zip(ids, band_structure))]
        band_structure = np.array(band_structure)

    if len(kk.shape) > 1:
        kkk = np.sum(kk, axis=0)
        if np.count_nonzero(kkk) == 1:
            kk = kk[:, np.where(kkk != 0.0)[0][0]]
        else:
            kk = list(range(kk.shape[0]))
            flag = False

    if show:

        vb = np.sort(np.real(band_structure))
        cb = vb.copy()
        vb[vb > 0] = np.NaN
        cb[cb < 0] = np.NaN

        vb = -np.sort(-vb)
        cb = np.sort(cb)

        try:
            fig, ax = plt.subplots(1, 1, figsize=(10, 5))
            if flag:
                fig.set_figheight(7)
                fig.set_figwidth(4)
                ax_lim1 = np.nanmax(vb) - 0.5
                ax_lim2 = np.nanmax(vb) + 0.1
                ax.set_ylim(ax_lim1, ax_lim2)
            ax.plot(kk, vb, marker="o", markersize=5)
            ax.set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
            ax.set_ylabel(r'Energy (eV)')
            ax.set_title('Valence band')
            fig.tight_layout()
            if show != 2:
                plt.show()
            else:
                plt.savefig(code_name + '_vb.pdf')
        except ValueError:
            pass

        try:
            fig, ax = plt.subplots(1, 1, figsize=(10, 5))
            if flag:
                fig.set_figheight(7)
                fig.set_figwidth(4)
                ax_lim1 = np.nanmin(cb) - 0.1
                ax_lim2 = np.nanmin(cb) + 0.7
                ax.set_ylim(ax_lim1, ax_lim2)
            ax.plot(kk, cb, marker="o", markersize=5)
            ax.set_xlabel(r'Wave vector ($\frac{\pi}{a}$)')
            ax.set_ylabel(r'Energy (eV)')
            ax.set_title('Conduction band')
            fig.tight_layout()
            if show != 2:
                plt.show()
            else:
                plt.savefig(code_name + '_cb.pdf')
        except ValueError:
            pass

    if save:
        with open(code_name+'.pkl', 'wb') as f:
            pickle.dump(band_structure, f, pickle.HIGHEST_PROTOCOL)


def main1(param_file, k_points_file, xyz, show, save, code_name):
    """

    Parameters
    ----------
    param_file :
        
    k_points_file :
        
    xyz :
        
    show :
        
    save :
        
    code_name :
        

    Returns
    -------

    """

    params, wave_vector, code_name = preprocess_data(param_file, k_points_file, xyz, code_name)

    # initialize Hamiltonian
    hamiltonian = tb.initializer(**params)

    # compute band structure
    band_structure = [{} for _ in range(len(wave_vector))]

    for j, jj in enumerate(wave_vector):
        vals, vects = hamiltonian.diagonalize_periodic_bc(jj)
        band_structure[j] = {'id': j, 'wave_vector': jj, 'eigenvalues': vals, 'eigenvectors': vects}
        print('#{} '.format(j), " ".join(['{:.3f} '.format(element) for element in vals]))

    postprocess_data(wave_vector, band_structure, show, save, code_name)

    return 0


def create_parser():
    """ """

    parser = argparse.ArgumentParser()

    parser.add_argument('param_file', type=str,
                        help='Path to the file in the yaml-format\
                             containing all parameters needed to run computations.')

    parser.add_argument('--k_points_file', type=str, default=None,
                        help='Path to the txt file containing coordinates of \
                             wave vectors for the band structure computations.\
                             If not specified, default values will be used.')

    parser.add_argument('--xyz', type=str, default=None,
                        help='Path to the file containing atomic coordinates. \
                             If specified, it overrides the coordinates \
                             specified in the param_files.')

    parser.add_argument('--show', '-S', type=int, default=1,
                        help='Show figures, 0/1/2. \
                             0 shows nothing,  \
                             1 outputs figures on screen, \
                             2 saves figures on disk without showing.')

    parser.add_argument('--save', '-s', type=int, default=0,
                        help='Save results of computations on disk, 0/1.')

    parser.add_argument('--code_name', type=str, default=None,
                        help='Code name is added to the names of all saved data files.')

    return parser


def main():
    """ """

    parser = create_parser()
    args = parser.parse_args()
    main1(args.param_file, args.k_points_file, args.xyz, args.show, args.save, args.code_name)

    return 0


if __name__ == '__main__':

    main()



import unittest
import nanonet.tb.tb_script as tb_script
import nanonet.tb.tbmpi_script as tbmpi_script


class Test(unittest.TestCase):
    """ """

    def setUp(self):
        """ """
        self.parser = tb_script.create_parser()

    def test_tb_without_mpi_sense(self):
        """ """

        args = self.parser.parse_args(['./examples/input_samples/input.yaml', '-S=0'])
        ans = tb_script.main1(args.param_file, args.k_points_file, args.xyz, args.show, args.save, args.code_name)
        self.assertEqual(ans, 0)

    def test_tb_without_mpi_sparse(self):
        """ """
        args = self.parser.parse_args(['./examples/input_samples/input.yaml', '-S=0'])
        ans = tbmpi_script.main1(args.param_file, args.k_points_file, args.xyz, args.show, args.save, args.code_name)
        self.assertEqual(ans, 0)


if __name__ == '__main__':

    unittest.main()




# from negf import pole_summation_method
#
#
# def test_density_integration():
#     """ """
#     ans = pole_summation_method()
#     ans1 = numerical_intrgration()
#     np.testing.assert_allclose(ans, ans1, rtol=1e-2)



import copy
import numpy as np
import scipy.linalg as linalg


def mat_left_div(mat_a, mat_b):
    """
    Solves the linear problem AX=B, where A is the linear operator,
    X is an unknown vector and B is the right-hand side column vector.

    Parameters
    ----------
    mat_a : numpy.ndarray (dtype=numpy.float)
        Linear operator
    mat_b : numpy.ndarray (dtype=numpy.float)
        Right-side column vector

    Returns
    -------
    ans : numpy.ndarray (dtype=numpy.float)
        Solution of the linear system
    """

    # ans, resid, rank, s = np.linalg.lstsq(mat_a, mat_b, rcond=1e-9)
    ans, resid, rank, s = linalg.lstsq(mat_a, mat_b, lapack_driver='gelsy')

    return ans


def _recursive_gf(energy, mat_l_list, mat_d_list, mat_u_list, s_in=0, s_out=0, damp=0.000001j):
    """The recursive Green's function algorithm is taken from
    M. P. Anantram, M. S. Lundstrom and D. E. Nikonov, Proceedings of the IEEE, 96, 1511 - 1550 (2008)
    DOI: 10.1109/JPROC.2008.927355
    
    In order to get the electron correlation function output, the parameters s_in has to be set.
    For the hole correlation function, the parameter s_out has to be set.

    Parameters
    ----------
    energy : numpy.ndarray (dtype=numpy.float)
        Energy array
    mat_d_list : list of numpy.ndarray (dtype=numpy.float)
        List of diagonal blocks
    mat_u_list : list of numpy.ndarray (dtype=numpy.float)
        List of upper-diagonal blocks
    mat_l_list : list of numpy.ndarray (dtype=numpy.float)
        List of lower-diagonal blocks
    s_in :
         (Default value = 0)
    s_out :
         (Default value = 0)
    damp :
         (Default value = 0.000001j)

    Returns
    -------
    g_trans : numpy.ndarray (dtype=numpy.complex)
        Blocks of the retarded Green's function responsible for transmission
    grd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    grl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gru : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gr_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    gnd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    gnl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gnu : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gin_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    gpd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    gpl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gpu : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gip_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    """
    # -------------------------------------------------------------------
    # ---------- convert input arrays to the matrix data type -----------
    # ----------------- in case they are not matrices -------------------
    # -------------------------------------------------------------------

    for jj, item in enumerate(mat_d_list):
        mat_d_list[jj] = item
        mat_d_list[jj] = mat_d_list[jj] - np.diag(energy * np.ones(mat_d_list[jj].shape[0]) + 1j*damp)

    # computes matrix sizes
    num_of_matrices = len(mat_d_list)  # Number of diagonal blocks.
    mat_shapes = [item.shape for item in mat_d_list]  # This gives the sizes of the diagonal matrices.

    # -------------------------------------------------------------------
    # -------------- compute retarded Green's function ------------------
    # -------------------------------------------------------------------

    # allocate empty lists of certain lengths
    gr_left = [None for _ in range(num_of_matrices)]
    gr_left[0] = mat_left_div(-mat_d_list[0], np.eye(mat_shapes[0][0]))  # Initialising the retarded left connected.

    for q in range(num_of_matrices - 1):  # Recursive algorithm (B2)
        gr_left[q + 1] = mat_left_div((-mat_d_list[q + 1] - mat_l_list[q].dot(gr_left[q]).dot(mat_u_list[q])),
                                      np.eye(mat_shapes[q + 1][0]))      # The left connected recursion.
    # -------------------------------------------------------------------

    grl = [None for _ in range(num_of_matrices-1)]
    gru = [None for _ in range(num_of_matrices-1)]
    grd = copy.copy(gr_left)                                             # Our glorious benefactor.
    g_trans = copy.copy(gr_left[len(gr_left)-1])

    for q in range(num_of_matrices - 2, -1, -1):                         # Recursive algorithm
        grl[q] = grd[q + 1].dot(mat_l_list[q]).dot(gr_left[q])           # (B5) We get the off-diagonal blocks for free.
        gru[q] = gr_left[q].dot(mat_u_list[q]).dot(grd[q + 1])           # (B6) because we need .Tthem.T for the next calc:
        grd[q] = gr_left[q] + gr_left[q].dot(mat_u_list[q]).dot(grl[q])  # (B4) I suppose I could also use the lower.
        g_trans = gr_left[q].dot(mat_u_list[q]).dot(g_trans)

    # -------------------------------------------------------------------
    # ------ compute the electron correlation function if needed --------
    # -------------------------------------------------------------------

    if isinstance(s_in, list):

        gin_left = [None for _ in range(num_of_matrices)]
        gin_left[0] = gr_left[0].dot(s_in[0]).dot(np.conj(gr_left[0]))

        for q in range(num_of_matrices - 1):
            sla2 = mat_l_list[q].dot(gin_left[q]).dot(np.conj(mat_u_list[q]))
            prom = s_in[q + 1] + sla2
            gin_left[q + 1] = np.real(gr_left[q + 1].dot(prom).dot(np.conj(gr_left[q + 1])))

        # ---------------------------------------------------------------

        gnl = [None for _ in range(num_of_matrices - 1)]
        gnu = [None for _ in range(num_of_matrices - 1)]
        gnd = copy.copy(gin_left)

        for q in range(num_of_matrices - 2, -1, -1):               # Recursive algorithm
            gnl[q] = grd[q + 1].dot(mat_l_list[q]).dot(gin_left[q]) +\
                     gnd[q + 1].dot(np.conj(mat_l_list[q])).dot(np.conj(gr_left[q]))
            gnd[q] = np.real(gin_left[q] +
                             gr_left[q].dot(mat_u_list[q]).dot(gnd[q + 1]).dot(np.conj(mat_l_list[q])).dot(np.conj(gr_left[q])) +
                             (gin_left[q].dot(np.conj(mat_u_list[q])).dot(np.conj(grl[q])) + gru[q].dot(mat_l_list[q]).dot(gin_left[q])))

            gnu[q] = np.conj(gnl[q])

    # -------------------------------------------------------------------
    # -------- compute the hole correlation function if needed ----------
    # -------------------------------------------------------------------
    if isinstance(s_out, list):

        gip_left = [None for _ in range(num_of_matrices)]
        gip_left[0] = gr_left[0].dot(s_out[0]).dot(np.conj(gr_left[0]))

        for q in range(num_of_matrices - 1):
            sla2 = mat_l_list[q].dot(gip_left[q]).dot(np.conj(mat_u_list[q]))
            prom = s_out[q + 1] + sla2
            gip_left[q + 1] = np.real(gr_left[q + 1].dot(prom).dot(np.conj(gr_left[q + 1])))

        # ---------------------------------------------------------------

        gpl = [None for _ in range(num_of_matrices - 1)]
        gpu = [None for _ in range(num_of_matrices - 1)]
        gpd = copy.copy(gip_left)

        for q in range(num_of_matrices - 2, -1, -1):               # Recursive algorithm
            gpl[q] = grd[q + 1].dot(mat_l_list[q]).dot(gip_left[q]) + gpd[q + 1].dot(np.conj(mat_l_list[q])).dot(np.conj(gr_left[q]))
            gpd[q] = np.real(gip_left[q] +
                             gr_left[q].dot(mat_u_list[q]).dot(gpd[q + 1]).dot(np.conj(mat_l_list[q])).dot(np.conj(gr_left[q])) +
                             gip_left[q].dot(np.conj(mat_u_list[q])).dot(np.conj(grl[q])) + gru[q].dot(mat_l_list[q]).dot(gip_left[q]))

            gpu[0] = gpl[0].conj().T

    # -------------------------------------------------------------------
    # -- remove energy from the main diagonal of th Hamiltonian matrix --
    # -------------------------------------------------------------------

    for jj, item in enumerate(mat_d_list):
        mat_d_list[jj] = mat_d_list[jj] + np.diag(energy * np.ones(mat_d_list[jj].shape[0]) + 1j*damp)

    # -------------------------------------------------------------------
    # ---- choose a proper output depending on the list of arguments ----
    # -------------------------------------------------------------------

    if not isinstance(s_in, list) and not isinstance(s_out, list):
        return g_trans, \
               grd, grl, gru, gr_left

    elif isinstance(s_in, list) and not isinstance(s_out, list):
        return g_trans, \
               grd, grl, gru, gr_left, \
               gnd, gnl, gnu, gin_left

    elif not isinstance(s_in, list) and isinstance(s_out, list):
        return g_trans, \
               grd, grl, gru, gr_left, \
               gpd, gpl, gpu, gip_left

    else:
        return g_trans, \
               grd, grl, gru, gr_left, \
               gnd, gnl, gnu, gin_left, \
               gpd, gpl, gpu, gip_left


def recursive_gf(energy, mat_l_list, mat_d_list, mat_u_list, left_se=None, right_se=None, s_in=0, s_out=0, damp=0.000001j):
    """The recursive Green's function algorithm is taken from
    M. P. Anantram, M. S. Lundstrom and D. E. Nikonov, Proceedings of the IEEE, 96, 1511 - 1550 (2008)
    DOI: 10.1109/JPROC.2008.927355

    In order to get the electron correlation function output, the parameters s_in has to be set.
    For the hole correlation function, the parameter s_out has to be set.

    Parameters
    ----------
    energy : numpy.ndarray (dtype=numpy.float)
        Energy array
    mat_d_list : list of numpy.ndarray (dtype=numpy.float)
        List of diagonal blocks
    mat_u_list : list of numpy.ndarray (dtype=numpy.float)
        List of upper-diagonal blocks
    mat_l_list : list of numpy.ndarray (dtype=numpy.float)
        List of lower-diagonal blocks
    s_in :
         (Default value = 0)
    s_out :
         (Default value = 0)
    damp :
         (Default value = 0.000001j)

    Returns
    -------
    g_trans : numpy.ndarray (dtype=numpy.complex)
        Blocks of the retarded Green's function responsible for transmission
    grd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    grl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gru : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gr_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    gnd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    gnl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gnu : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gin_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    gpd : numpy.ndarray (dtype=numpy.complex)
        Diagonal blocks of the retarded Green's function
    gpl : numpy.ndarray (dtype=numpy.complex)
        Lower diagonal blocks of the retarded Green's function
    gpu : numpy.ndarray (dtype=numpy.complex)
        Upper diagonal blocks of the retarded Green's function
    gip_left : numpy.ndarray (dtype=numpy.complex)
        Left-conencted blocks of the retarded Green's function
    """

    if isinstance(left_se, np.ndarray):
        s01, s02 = mat_d_list[0].shape
        left_se = left_se[:s01, :s02]
        mat_d_list[0] = mat_d_list[0] + left_se

    if isinstance(right_se, np.ndarray):
        s11, s12 = mat_d_list[-1].shape
        right_se = right_se[-s11:, -s12:]
        mat_d_list[-1] = mat_d_list[-1] + right_se

    ans = _recursive_gf(energy, mat_l_list, mat_d_list, mat_u_list, s_in=s_in, s_out=s_out, damp=damp)

    if isinstance(left_se, np.ndarray):
        mat_d_list[0] = mat_d_list[0] - left_se

    if isinstance(right_se, np.ndarray):
        mat_d_list[-1] = mat_d_list[-1] - right_se

    return ans


from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('requirements.txt') as fp:
    install_requires = fp.read().splitlines()

setup(name='nano-net',
      version='1.3.6',
      description='Python framework for tight-binding computations',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='M. V. Klymenko, J. A. Vaitkus, J. S. Smith, J. H. Cole',
      author_email='mike.klymenko@rmit.edu.au',
      license='MIT',
      packages=find_namespace_packages(include=['nanonet.*']),
      entry_points={
        'console_scripts': ['tb = tb.tb_script:main', 'tbmpi = tb.tbmpi_script:main', 'gf = tb.gf_script:main'],
      },
      url = "https://github.com/freude/NanoNet",
      zip_safe=False,
      classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
      ],
      python_requires='>=3.6',
      install_requires=install_requires
      )



import numpy as np
from scipy.interpolate import RegularGridInterpolator
import matplotlib.pyplot as plt
import matplotlib.lines as mlines
from matplotlib import rc


def _getline(cube):
    """Read a line from cube file where first field is an int
    and the remaining fields are floats.

    Parameters
    ----------
    cube :
        file object of the cube file

    Returns
    -------
    type
        

    """
    line = cube.readline().strip().split()
    return int(line[0]), np.array(list(map(float, line[1:])))


def read_cube(fname):
    """Read cube file into numpy array

    Parameters
    ----------
    fname :
        filename of cube file

    Returns
    -------
    type
        

    """
    bohr = 0.529177
    meta = {}
    with open(fname, 'r') as cube:
        cube.readline()
        cube.readline()  # ignore comments
        natm, meta['org'] = _getline(cube)
        nx, meta['xvec'] = _getline(cube)
        ny, meta['yvec'] = _getline(cube)
        nz, meta['zvec'] = _getline(cube)

        if nx > 0:
            meta['xvec'] = meta['xvec'] * bohr
        if ny > 0:
            meta['yvec'] = meta['yvec'] * bohr
        if nz > 0:
            meta['zvec'] = meta['zvec'] * bohr

        meta['atoms'] = [_getline(cube) for i in range(natm)]
        data = np.zeros((nx * ny * nz))
        idx = 0
        for line in cube:
            for val in line.strip().split():
                data[idx] = float(val)
                idx += 1

    data = np.reshape(data, (nx, ny, nz))
    return data, meta


class Field(object):
    """ """

    def __init__(self, path):

        self._origin_shift = np.array([0.0, 0.0, 0.0])
        self._cube = []
        self._origin_has_changed = False
        self._rot_mat = []
        self._atoms = []

        if path.endswith('.cube') or path.endswith('.cub'):
            data, meta = read_cube(path)
            data = np.swapaxes(data, 1, 2)

            shape = data.shape
            x = np.linspace(0.0, meta['xvec'][0] * shape[0], shape[0]) - 0.5 * meta['xvec'][0] * shape[0]
            z = np.linspace(0.0, meta['yvec'][1] * shape[1], shape[1]) - 0.5 * meta['yvec'][1] * shape[1]
            y = np.linspace(0.0, meta['zvec'][2] * shape[2], shape[2]) - 0.5 * meta['zvec'][2] * shape[2]

            self._cube = [[x[0], y[0], z[0]], [x[-1], y[-1], z[-1]]]

            self.origin = (0, 0, 0)
            self._shape = shape
            self._atoms = meta['atoms']

        data[data > 100] = 100
        data[data < -100] = -100
        self._interpolant = RegularGridInterpolator((x, y, z), data, bounds_error=False)

    def set_origin(self, origin):
        """Set the coordinates of the center of the molecule

        Parameters
        ----------
        origin :
            return:

        Returns
        -------

        """

        if isinstance(origin, list):
            origin = np.array(origin)

        self._origin_shift = origin
        self._origin_has_changed = True

    def rotate(self, axis, theta):
        """Set the coordinates of the center of the molecule

        Parameters
        ----------
        origin :
            return:
        axis :
            
        theta :
            

        Returns
        -------

        """
        if axis == 'x':
            rot_mat = np.array([[1.0, 0.0, 0.0],
                                [0.0, np.cos(theta), -np.sin(theta)],
                                [0.0, np.sin(theta), np.cos(theta)]])
        elif axis == 'y':
            rot_mat = np.array([[np.cos(theta), 0.0, np.sin(theta)],
                                [0.0, 1.0, 0.0],
                                [-np.sin(theta), 0.0, np.cos(theta)]])
        elif axis == 'z':
            rot_mat = np.array([[np.cos(theta), -np.sin(theta), 0.0],
                                [np.sin(theta), np.cos(theta), 0.0],
                                [0.0, 0.0, 1.0]])
        else:
            raise ValueError('Wrong axis')

        self._rot_mat.append(rot_mat)

    def reset_rotations(self):
        """ """

        self._rot_mat = []

    def _transform(self, coords1, translate):
        """

        Parameters
        ----------
        coords1 :
            
        translate :
            

        Returns
        -------

        """

        coords = coords1.copy()

        if len(coords.shape) < 2:
            coords = [coords]

        for j in range(len(coords)):
            coords[j] += self.origin

            if self._origin_has_changed:
                coords[j] = coords[j] - self._origin_shift

            if isinstance(translate, np.ndarray):
                coords[j] = coords[j] - np.squeeze(translate)

            if len(self._rot_mat) > 0:
                for item in self._rot_mat:
                    coords[j] = np.dot(item, coords[j].T).T[0]

        return coords

    def _inv_transform(self, coords1, translate):
        """

        Parameters
        ----------
        coords1 :
            
        translate :
            

        Returns
        -------

        """

        coords = coords1.copy()

        if len(coords.shape) < 2:
            coords = [coords]

        for j in range(len(coords)):

            if len(self._rot_mat) > 0:
                for item in reversed(self._rot_mat):
                    coords[j] = np.dot(np.linalg.inv(item), coords[j].T).T[0]

            if isinstance(translate, np.ndarray):
                coords[j] = coords[j] + np.squeeze(translate)

            if self._origin_has_changed:
                coords[j] = coords[j] + self._origin_shift

            coords[j] -= self.origin

        return coords

    def get_values(self, coords1, translate=None):
        """

        Parameters
        ----------
        coords1 :
            
        translate :
             (Default value = None)

        Returns
        -------

        """

        coords = self._transform(coords1, translate)
        values = self._interpolant(coords)

        return np.nan_to_num(values)

    def get_atoms(self):
        """ """

        ind = []
        coords = []

        for item in self._atoms:
            ind.append(item[0])
            coords.append(item[1][1:])

        return np.array(ind), np.array(coords)

    def add_screening(self, eps, mol_y_length, spacing):
        """

        Parameters
        ----------
        eps :
            
        mol_y_length :
            
        spacing :
            

        Returns
        -------

        """

        # from negf.poisson import laplacian, PoissonSolver

        num_points = 256

        x = np.linspace(-19, 19, num_points - 1)
        y = np.linspace(-19, 19, num_points - 1)
        z = np.linspace(-19, 19, num_points - 1)

        X = np.meshgrid(x, y, z)

        coords = np.vstack((X[0].flatten(), X[1].flatten(), X[2].flatten())).T

        vec = np.zeros(3)
        eps[1] = mol_y_length * 0.5
        vec[eps[0]] = 1.0
        vec = np.array(vec)

        rot_mat = np.identity(3)
        # rot_mat[:2, :2] = np.matrix(np.sqrt(1.0 - self._rot_mat[0][1:, 1:]**2))
        rot_mat[:2, :2] = self._rot_mat[0][1:, 1:]
        rot_mat[0, 1] = -rot_mat[0, 1]
        rot_mat[1, 0] = -rot_mat[1, 0]
        # rot_mat = np.linalg.inv(rot_mat)
        # rot_mat = np.sqrt(1.0 - rot_mat ** 2)
        # rot_mat[2, 2] = 1.0
        # rot_mat[0, 0] = -rot_mat[0, 0]
        # rot_mat[1, 1] = -rot_mat[1, 1]

        vec = np.dot(rot_mat, vec.T).T[0]

        dist = X[0] * vec[0] + X[1] * vec[1] + X[2] * vec[2] + eps[1] + \
               0.5 * mol_y_length * (1.0 - np.sin(1.13446)) + \
               spacing

        perm = eps[2] * (1.0 - 0.5 * np.tanh(eps[1] * dist) - 0.5) \
               + eps[3] * (0.5 * np.tanh(eps[1] * dist) + 0.5)

        # perm = np.ones(perm.shape)

        pot = self._interpolant(np.vstack((X[0].flatten(), X[1].flatten(), X[2].flatten())).T).reshape(X[0].shape)
        pot1 = pot / eps
        # dens = -laplacian(pot, x[2] - x[1]) / (4 * np.pi)

        # ps = PoissonSolver()
        # ps.solve_mesh(X[0], X[1], X[2], pot1, dens, perm)

        # pot1 = pot1 + 0.1 * perm / np.max(perm)

        # pot1 = np.swapaxes(pot1, 1, 0)
        self._interpolant = RegularGridInterpolator((x, y, z), pot1, bounds_error=False)


class Field1D(object):
    """ """

    def __init__(self, coord_dependence, axis=2):

        self._func = coord_dependence
        self._axis = axis

    def get_values(self, coords1, translate=None):
        """

        Parameters
        ----------
        coords1 :
            
        translate :
             (Default value = None)

        Returns
        -------

        """

        coords = coords1.copy()
        values = np.zeros(len(coords))

        if len(coords.shape) < 2:
            coords = [coords]

        for j in range(len(coords)):
            if isinstance(translate, np.ndarray):
                coords[j] -= translate

            values[j] = self._func(coords[j][self._axis])

        return values


def main():
    """ """
    import matplotlib.pyplot as plt

    fl = Field(path='/home/mk/gaussian_swarm/gauss_comp/out_neutral.cube')

    x = np.linspace(fl._cube[0][0], fl._cube[1][0], fl._shape[0])
    y = np.linspace(fl._cube[0][1], fl._cube[1][1], fl._shape[1])
    z = np.linspace(fl._cube[0][2], fl._cube[1][2], fl._shape[2])

    x = np.linspace(0, 60, 100)
    y = np.linspace(0, 11, 100)
    z = np.linspace(0, 5.5, 100)

    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    # data = fl.get_values(np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T, translate=np.array([0,0,15]))
    # data = data.reshape(X.shape)
    # plt.imshow(data[:, :, 50])
    # # plt.contour(data[:, 50, :], levels=(0.04, 0.06, 0.08, 0.1, 0.2))
    # plt.show()

    fl.set_origin(np.array([20, 11, 2.75]))

    data = fl.get_values(np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T)
    data = data.reshape(X.shape)
    plt.imshow(data[:, :, 50])

    # plt.contour(data[:, 50, :], levels=(0.04, 0.06, 0.08, 0.1, 0.2))

    plt.show()


def laplacian(input_mat, dx):
    """

    Parameters
    ----------
    input_mat :
        
    dx :
        

    Returns
    -------

    """
    top = input_mat[0:-2, 1:-1, 1:-1]
    left = input_mat[1:-1, 0:-2, 1:-1]
    bottom = input_mat[2:, 1:-1, 1:-1]
    right = input_mat[1:-1, 2:, 1:-1]
    center = input_mat[1:-1, 1:-1, 1:-1]
    back = input_mat[1:-1, 1:-1, 0:-2]
    forward = input_mat[1:-1, 1:-1, 2:]
    return np.pad(top + left + bottom + right + back + forward - 6 * center,
                  ((1, 1), (1, 1), (1, 1)),
                  'constant',
                  constant_values=((0, 0), (0, 0), (0, 0))) / dx ** 2


def main1():
    """ """
    import matplotlib.pyplot as plt

    # fl = Field(path='/home/mk/gaussian_swarm/gauss_comp/wB_ion.cube')
    # fl = Field(path='../cubefil.cube')
    # fl1 = Field(path='/home/mk/gaussian_swarm/gauss_comp/cam_ion.cube')

    fl = Field(path='/home/mk/tetracene_dft_wB_pcm_38_32.cube')
    fl1 = Field(path='/home/mk/tetracene_dft_wB_pcm_38_32_wpcm.cube')

    x = np.linspace(fl._cube[0][0], fl._cube[1][0], fl._shape[0])
    y = np.linspace(fl._cube[0][1], fl._cube[1][1], fl._shape[1])
    z = np.linspace(fl._cube[0][2], fl._cube[1][2], fl._shape[2])

    x = np.linspace(-30, 30, 200)
    y = np.linspace(-30, 30, 200)
    z = np.linspace(-30, 30, 200)

    X, Y, Z = np.meshgrid(x, y, z, indexing='ij')

    # data = fl.get_values(np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T, translate=np.array([0,0,15]))
    # data = data.reshape(X.shape)
    # plt.imshow(data[:, :, 50])
    # # plt.contour(data[:, 50, :], levels=(0.04, 0.06, 0.08, 0.1, 0.2))
    # plt.show()

    # fl.set_origin(np.array([0.0, 0.0, 1.0]))
    # fl1.set_origin(np.array([0.0, 0.0, 1.0]))

    fl.set_origin(np.array([6.36, 11.86, 2.75]))
    fl.rotate('z', 1.13446)  # 65 degrees
    data = fl.get_values(np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T)
    data = data.reshape(X.shape)

    data1 = fl1.get_values(np.vstack((X.flatten(), Y.flatten(), Z.flatten())).T)
    data1 = data1.reshape(X.shape)

    cut_level = 0.015

    data1[data1 > cut_level] = cut_level
    data[data > cut_level] = cut_level

    from matplotlib import cm
    norm = cm.colors.Normalize(vmax=cut_level, vmin=-cut_level)
    # cmap = cm.PRGn
    # cmap = cm.seismic
    cmap = cm.coolwarm
    levels = np.arange(-cut_level * 1.1, cut_level * 1.1, 0.002)
    cset = plt.contourf(X[:, :, 100], Y[:, :, 100], data[:, :, 100], levels, norm=norm,
                        cmap=cm.get_cmap(cmap, len(levels) - 1))

    from matplotlib.patches import Rectangle
    currentAxis = plt.gca()
    someX = -5
    someY = -5
    width = 10
    height = 10
    currentAxis.add_patch(Rectangle((someX, someY), width, height,
                                    alpha=1, fill=None))

    plt.show()

    print('hi')


# class nf(float):
#     def __repr__(self):
#         str = '%.3f eV' % (self.__float__(),)
#         if str[-1] == '0':
#             return '%.2f eV' % self.__float__()
#         else:
#             return '%.3f eV' % self.__float__()


class nf(float):
    """ """
    def __repr__(self):
        if self.__float__() == 0.0:
            return '0.0'
        else:
            return '%.1e eV' % (self.__float__(),)


def plot(x, y, data, data1, cut_level, n_contours):
    """

    Parameters
    ----------
    x :
        
    y :
        
    data :
        
    data1 :
        
    cut_level :
        
    n_contours :
        

    Returns
    -------

    """
    X, Y = np.meshgrid(x, y, indexing='ij')

    data[data > cut_level] = cut_level

    from matplotlib import cm
    norm = cm.colors.Normalize(vmax=cut_level, vmin=-cut_level)
    # cmap = cm.PRGn
    # cmap = cm.seismic
    cmap = cm.coolwarm
    levels = np.arange(-cut_level * 1.1, cut_level * 1.1, 2 * cut_level / n_contours)
    # levels = np.linspace(-0.1, 0.1, 21)
    # levels = np.linspace(-0.01, 0.01, 21)
    fig, ax = plt.subplots(figsize=(10, 10))
    cset = ax.contourf(X, Y, data, levels, norm=norm,
                       cmap=cm.get_cmap(cmap, len(levels) - 1))
    cset = ax.contour(X, Y, data, levels, norm=norm,
                      colors='r', linewidths=1)

    cset = ax.contour(X, Y, data1, levels, norm=norm, colors='k')

    # Recast levels to new class
    cset.levels = [nf(val) for val in cset.levels]

    # Label levels with specially formatted floats

    fmt = '%r'

    ax.clabel(cset, cset.levels[9:-9], inline=True, fmt=fmt, fontsize=14)
    # ax.clabel(cset, cset.levels, inline=True, fmt=fmt, fontsize=14)
    ax.axis('off')
    l = mlines.Line2D([10, 15], [-15, -15], linewidth=3, color='k')
    ax.add_line(l)
    rc('font', **{'family': 'monospace', 'weight': 'bold'})
    rc('text', usetex=True)
    ax.text(12, -14.4, r'$5 \mathring{A}$', fontsize=16)


if __name__ == '__main__':
    main1()

    # path = '/home/mk/tetracene_dft_wB_pcm_38_32.cube'
    # data, meta = read_cube(path)
    # path = '/home/mk/tetracene_dft_wB_pcm_38_32_wpcm.cube'
    # data1, meta = read_cube(path)
    # x = np.linspace(0.0, meta['xvec'][0] * data.shape[0], data.shape[0]) - 0.5 * meta['xvec'][0] * data.shape[0]
    # data = data.reshape((len(x), len(x), len(x)))
    # cut_level = 0.015
    # cut_level = 0.01
    # plot(x, x, data[:, 100, :] / 3.8, data1[:, 100, :] / 3.8, cut_level, 20)
    # # plot(x, x, data1[:, 100, :] / 3.8, cut_level, 20)
    # plt.show()



import numpy as np
from scipy.linalg import block_diag
from nanonet.negf.field import Field
from nanonet.tb.aux_functions import fd


class HamiltonianChain(object):
    """ """

    def __init__(self, h_l, h_0, h_r, coords):

        self.h_l = h_l
        self.h_0 = h_0
        self.h_r = h_r
        self._coords = coords
        self._z_coords_map = []

        self.num_sites = h_0.shape[0]

        self.elem_length = None
        self.left_translations = None
        self.right_translations = None
        self.fields = None
        self.sgf_l = None
        self.sgf_r = None

        self.energy = 0
        self.tempr = 0
        self.ef1 = 0
        self.ef2 = 0

    @property
    def sgf(self):
        """ """

        sgf = [None for _ in range(len(self.h_0))]

        for jjj in range(len(self.h_0)):
            if jjj == 0:
                sgf[jjj] = -2.0 * np.imag(self.sgf_l) * fd(self.energy, self.ef1, self.tempr)
            elif jjj == len(self.h_0) - 1:
                sgf[jjj] = -2.0 * np.imag(self.sgf_r) * fd(self.energy, self.ef2, self.tempr)
            else:
                sgf[jjj] = np.zeros(self.h_0[jjj].shape)

        return sgf

    def translate(self, period, left_translations, right_translations):
        """

        Parameters
        ----------
        period :
            
        left_translations :
            
        right_translations :
            

        Returns
        -------

        """

        self.elem_length = np.array(period)
        self.left_translations = left_translations
        self.right_translations = right_translations

        self.h_l = [self.h_l for _ in range(left_translations)] + \
                   [self.h_l for _ in range(right_translations)]

        self.h_0 = [self.h_0 for _ in range(left_translations)] + \
                   [self.h_0] + \
                   [self.h_0 for _ in range(right_translations)]

        self.h_r = [self.h_r for _ in range(left_translations)] + \
                   [self.h_r for _ in range(right_translations)]

    def add_field(self, field, eps=7.0):
        """

        Parameters
        ----------
        field :
            
        eps :
             (Default value = 7.0)

        Returns
        -------

        """

        field_buf = []

        for jjj in range(self.left_translations, 0, -1):
            field_buf.append(field.get_values(self._coords, translate=jjj * self.elem_length) / eps)

        field_buf.append(field.get_values(self._coords) / eps)

        for jjj in range(1, self.right_translations + 1):
            field_buf.append(field.get_values(self._coords, translate=-jjj * self.elem_length) / eps)

        field_buf = np.array(field_buf)

        if not isinstance(self.fields, list):
            self.fields = []

        self.fields.append(field_buf)

        for jjj in range(len(self.h_0)):
            self.h_0[jjj] = self.h_0[jjj] - np.diag(field_buf[jjj])

    def remove_field(self):
        """ """

        if isinstance(self.fields, list):
            for item in self.fields:
                for jjj in range(len(self.h_0)):
                    self.h_0[jjj] = self.h_0[jjj] - np.diag(item[jjj])

        self.fields = None

    def add_self_energies(self, sgf_l, sgf_r, energy=0, tempr=0, ef1=0, ef2=0):
        """

        Parameters
        ----------
        sgf_l :
            
        sgf_r :
            
        energy :
             (Default value = 0)
        tempr :
             (Default value = 0)
        ef1 :
             (Default value = 0)
        ef2 :
             (Default value = 0)

        Returns
        -------

        """

        self.energy = energy
        self.tempr = tempr
        self.ef1 = ef1
        self.ef2 = ef2

        self.sgf_l = sgf_l
        self.sgf_r = sgf_r

        self.h_0[-1] = self.h_0[-1] + sgf_r
        self.h_0[0] = self.h_0[0] + sgf_l

    def remove_self_energies(self):
        """ """

        self.h_0[-1] = self.h_0[-1] - self.sgf_r
        self.h_0[0] = self.h_0[0] - self.sgf_l

        self.sgf_l = None
        self.sgf_r = None

    def translate_self_energies(self, sgf_l, sgf_r):
        """

        Parameters
        ----------
        sgf_l :
            
        sgf_r :
            

        Returns
        -------

        """

        mat_list = [item * 0.0 for item in self.h_0]

        return block_diag(*tuple([sgf_r] + mat_list[1:])), block_diag(*tuple(mat_list[:-1] + [sgf_l]))

    @property
    def coords(self):
        """ """
        if self.elem_length is not None:
            coords = [self._coords - jjj * self.elem_length for jjj in range(self.left_translations, 0, -1)] + \
                     [self._coords] + \
                     [self._coords + jjj * self.elem_length for jjj in range(1, self.right_translations + 1)]

            return np.concatenate(coords)
        else:
            return self._coords

    def z_coords_map(self, ind):
        """

        Parameters
        ----------
        ind :
            

        Returns
        -------

        """

        if len(self._z_coords_map) == 0:

            coords = self.coords[:, 2]
            unique_coords = self.z_coords()
            coord_map = np.zeros(coords.shape)

            for j, item in enumerate(unique_coords):
                coord_map[np.abs(coords-item) < 0.001] = j

            self._z_coords_map = coord_map

        return int(self._z_coords_map[ind])

    def z_coords(self):
        """ """

        unique_coords = np.sort(np.array(list(set(self.coords[:, 2]))))

        return unique_coords

    def get_matrix(self):
        """ """

        if isinstance(self.h_0, list):
            matrix = block_diag(*tuple(self.h_0))
        else:
            return self.h_0

        for j in range(len(self.h_l)):

            s1, s2 = self.h_0[j].shape

            matrix[j * s1:(j + 1) * s1, (j + 1) * s2:(j + 2) * s2] = self.h_r[j]
            matrix[(j + 1) * s1:(j + 2) * s1, j * s2:(j + 1) * s2] = self.h_l[j]

        return matrix

    def visualize(self):
        """ """

        import matplotlib.pyplot as plt
        from matplotlib import cm

        vals = np.zeros(len(self.fields[0])*len(self.fields[0][0]))

        for field in self.fields:
            vals += np.hstack(tuple([item for item in field]))

        vals /= np.max(vals)

        from mpl_toolkits.mplot3d import Axes3D
        fig = plt.figure()
        ax = fig.add_subplot(111, projection='3d')
        norm = cm.colors.Normalize(vmax=np.abs(np.max(vals)), vmin=-np.abs(np.min(vals)))

        ax.scatter(self.coords[:, 0], self.coords[:, 1], self.coords[:, 2],
                   c=vals,
                   marker='o',
                   norm=norm,
                   cmap=cm.get_cmap(cm.coolwarm))

        plt.show()


class HamiltonianChainComposer(HamiltonianChain):
    """ """

    def __init__(self, h_l, h_0, h_r, coords, params):
        super(HamiltonianChainComposer, self).__init__(h_l, h_0, h_r, coords)

        self.translate(params['unit_cell'], params['left_translations'], params['right_translations'])
        self.mol_length = 0
        uc = np.array(params['unit_cell'])

        self.dict_of_fields = {}

        if 'fields' in params:

            for item in params['fields']['xyz']:

                field = list(item.keys())[0]

                if field in self.dict_of_fields:
                    self.dict_of_fields[field].set_origin(self.dict_of_fields[field].mean_coords + np.array(item[field]))
                else:
                    self.dict_of_fields[field] = Field(path=params['fields'][field])
                    # angle = 1.13446
                    angle = params['fields']['angle']

                    self.dict_of_fields[field].rotate('x', angle)  # 65 degrees
                    self.dict_of_fields[field].rotate('y', np.pi / 2.0)

                    size_x_max = np.max(coords[:, 0])
                    size_y_min = np.min(coords[:, 1])
                    size_y_max = np.max(coords[:, 1])

                    _, mol_coords = self.dict_of_fields[field].get_atoms()
                    mol_y_length0 = np.max(mol_coords[:, 1]) - np.min(mol_coords[:, 1])
                    mol_y_length = mol_y_length0 * np.sin(angle)
                    mol_z_length = mol_y_length0 * np.cos(angle)

                    self.dict_of_fields[field].mean_coords = np.array([0.5 * (size_x_max - np.abs(size_y_min)),
                                                                  size_y_max + 0.5 * mol_y_length +
                                                                  params['fields']['spacing'],
                                                                  0.5 * mol_z_length])

                    self.dict_of_fields[field].set_origin(self.dict_of_fields[field].mean_coords + np.array(item[field]))

                if isinstance(params['fields']['eps'], list):
                    self.dict_of_fields[field].add_screening(params['fields']['eps'], mol_y_length0, params['fields']['spacing'])
                    # self.add_field(self.dict_of_fields[field], eps=params['fields']['eps'][3])
                    self.add_field(self.dict_of_fields[field], eps=1.0)
                else:
                    self.add_field(self.dict_of_fields[field], eps=params['fields']['eps'])


if __name__ == '__main__':

    # fields_config = """
    #
    # unit_cell:        [[0, 0, 5.50]]
    #
    # left_translations:     3
    # right_translations:    3
    #
    # fields:
    #
    #     eps = 3.8
    #
    #     cation:      '/home/mk/tetracene_dft_wB_pcm_38_32_cation.cube'
    #
    #     angle:       1.13446
    #     spacing:     5.0
    #
    #     xyz:
    #         - cation:       [0.0000000000,    0.0000000000,    0.0000000000]
    #
    # """
    #
    # params = yaml_parser(fields_config)
    # print('hi')
    import numpy as np
    from matplotlib import pyplot

    fig = pyplot.figure()
    ax = fig.add_axes((0.1, 0.1, 0.8, 0.8), autoscale_on=False)

    # Empty data plot
    points = ax.scatter([], [], color='r', zorder=2)
    # ax properties
    ax.set_xlim(-10, 10)
    ax.set_ylim(5e-2, 5e3)
    ax.set_yscale("log")

    # Example data points
    x_data = [-5, -3, 0, 3, 5]
    y_data = [1, 10, 1000, 10, 1]
    # Set data points
    points.set_offsets(np.hstack((x_data, y_data)).T)

    pyplot.show()